/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * dialogs
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <strings.h>

#include "dialog.h"
#include "rli.h"
#include "common.h"
#include "screen.h"
#include "comkeys.h"
#include "list.h"
#include "and_keys.h"
#include "text.h"
#include "and_scr.h"
#include "list.h"
#include "windows.h"
#include "console.h"
#include "rlci.h"
#include "ezconfig.h"

#ifdef IBMPC
#define DBOX_ATTR   BLKGRY
#define DBOX_PAINT  WHTGRY
#define DBOX_HILITE WHTGRY
#define DBOX_BUTTON RVDGRY
#else /* IBMPC */
#define DBOX_ATTR   NORMAL
#define DBOX_PAINT  NORMAL
#define DBOX_HILITE HILITE
#define DBOX_BUTTON RVIDEO
#endif /* IBMPC */

#define START_TEXTROW  3
#define START_TEXTCOL  2
#define START_BUTTON   3

#define CANCEL_BUTTON   0x0001
#define SAVE_BUTTON     0x0002
#define OK_BUTTON       0x0004
#define NO_BUTTON       0x0008
#define YES_BUTTON      0x0010
#define SET_BUTTON      0x0020
#define CLEAR_BUTTON    0x0040
#define QUERY_BUTTON    0x0080
#define MORE_BUTTON     0x0100

typedef struct string_struct_R {
  POINT loc;
  int bufsize;
  int datalen;
  int scrsize;
  int curpos;
  int chrpos;
  int attr;
  char *buf;
  struct string_struct_R *next;
} STRINFO, *STRINFO_pt;

typedef struct button_struct_R {
  BOX bounds;
  char *text;
  int   key, mask;
  Boolean is_default;
  struct button_struct_R *next;
} BUTTON, *BUTTON_pt;

typedef struct checkbox_struct_R {
  POINT loc;
  Boolean val;
  struct checkbox_struct_R *next;
} CHECKBOX, *CHECKBOX_pt;

typedef struct static_text_R {
  BOX bounds;
  char *text;
  int   attr;
  struct static_text_R *next;
} STEXT, *STEXT_pt;

typedef struct dialog_R {
  BOX bounds;
  char *title;
  STEXT_pt stx;
  BUTTON_pt bts;
  CHECKBOX_pt cbx, cur_cbx;
  STRINFO_pt sfo,  cur_sfo;
} DIALOG, *DIALOG_pt;

extern char *getvstr(), *SCR_GetKeyText();

char *login_prompts[] = {
  "Userid:  ",
  "IMAP Server:",
  "SMTP Server:",
};

char *standard_title = "Press RETURN for Ok, %s to Cancel";

DIALOG_pt gl_adb = (DIALOG_pt)NULL;

char *about_text[] = {
"EEEEEE   ZZZZZZZ   I   MM      MM    AAA    I   L    ",
"E             Z    I   M M    M M   A   A   I   L    ",
"EEE        ZZZ     I   M  M  M  M   AAAAA   I   L    ",
"E         Z        I   M   MM   M   A   A   I   L    ",
"EEEEEE   ZZZZZZZ   I   M        M   A   A   I   LLLLL",
"                                                     ",
"This product includes software developed by Computing",
"       Services at Carnegie Mellon University        ",
"           (http://www.cmu.edu/computing/)           "
};

/* Forward declarations */
void StripWhiteEnds(char *buf);
void DLG_FormatText(char *text,char *buf,int w);
void DLG_Box(BOX *adb,char *title);
int DLG_ShiftData(STRINFO *asi,int delta);
int DLG_ValidKey(int key);
int DLG_InsertChar(STRINFO *asi,int key);
void DLG_Write(char *text,int r,int c,int w,int attr);
void DLG_ShowText(BOX *adb,char *text,int size, int attr);
void DLG_SetCursor(int r,int c);
void DLG_DrawStrField(STRINFO_pt sfo);
void DLG_DrawCheckBox(CHECKBOX_pt cbx);
void DLG_DrawButton(BUTTON_pt abutton,int attr);
void DLG_InitCheckBox(CHECKBOX_pt cbx,int r,int c,int val);
void DLG_MakeButton(BUTTON_pt bts,int r, int c, int w,		    
		    char *name,int key, int mask, int is_default);
void DLG_FlashButton(BUTTON *abutton);
void DLG_SetupButtons(POINT *dim,BUTTON_pt bts,
		      int num_bts, int bts_mask, int default_mask);
void DLG_InitStaticText(STEXT_pt stx,int r,int c,int h,int w,char *text,int attr);
int DLG_ProcessKey(STRINFO *asi,int key);
int DLG_CheckButton(int key,int check_first_char);
int DLG_GetCheckBox(CHECKBOX_pt cbx);
int DLG_GetStrField(STRINFO *asi);
void DLG_InitStrInfo(STRINFO *asi,char *buf,int bufsize, int row, int col, int w, int attr);

int DLG_FaxHeadersOptions(void);


void DLG_DrawAllStaticText(void)
{
  STEXT_pt stx = gl_adb->stx;

  while(stx) {
    BOX *bounds = &stx->bounds;
    DLG_ShowText(bounds, stx->text,
		 (int)(bounds->dim.x * bounds->dim.y), stx->attr);
    stx = stx->next;
  }
}

void DLG_DrawAllButtons(void)
{
  BUTTON_pt bts = gl_adb->bts;

  while(bts) {
    DLG_DrawButton(bts, DBOX_BUTTON);
    bts = bts->next;
  }
}

void DLG_DrawAllCheckBoxes(void)
{
  CHECKBOX_pt cbx = gl_adb->cbx;

  while (cbx) {
      DLG_DrawCheckBox(cbx);
      cbx = cbx->next;
  }
}

void DLG_DrawAllStrFields(void)
{
  STRINFO_pt sfo = gl_adb->sfo;

  while (sfo) {
      DLG_DrawStrField(sfo);
      sfo = sfo->next;
  }
}

void DLG_DrawContent(void)
{
  DLG_DrawAllStaticText();
  DLG_DrawAllButtons();
  DLG_DrawAllCheckBoxes();
  DLG_DrawAllStrFields();
  SCR_UpdateScreen();
}

void DLG_Draw(void)
{
  DLG_Box(&gl_adb->bounds, gl_adb->title);
  DLG_DrawContent();
}

DIALOG_pt DLG_Setup(DIALOG_pt dlog,int num_bts, int bts_mask, int bts_def)
{
  DIALOG_pt prev_dialog = gl_adb;

  dlog->bounds.loc.x = (screen_cols - dlog->bounds.dim.x) / 2;
  dlog->bounds.loc.y = (screen_rows - dlog->bounds.dim.y) / 2;

  SCR_CursorOff();

  if (prev_dialog == (DIALOG_pt)NULL) WIN_HLSelection(FALSE);

  gl_adb = dlog;

  if (num_bts && dlog->bts)
    DLG_SetupButtons(&dlog->bounds.dim, dlog->bts, num_bts, bts_mask, bts_def);
  else
    dlog->bts = (BUTTON_pt)NULL;

  DLG_Draw();

  return(prev_dialog);
}

int DLG_DoKey(int key)
{
  if (gl_adb->bts)
    return(DLG_CheckButton(key, (gl_adb->sfo == NULL)));
  else {
      if ((key == KEYCODE_RETURN) || (key == KEYCODE_FAKERETURN))
	  return OK_BUTTON;
      else
	  return 0;
  }
}

int DLG_NextField(int key,CHECKBOX_pt *cbx,STRINFO_pt  *sfo)
{
  switch(key) {
  case KEYCODE_UP:
  case KEYCODE_DOWN:
    if (*cbx) {
      if (key == KEYCODE_DOWN) {
	*cbx = (*cbx)->next;
	if (!*cbx) *cbx = gl_adb->cbx;
      } else {
	CHECKBOX_pt tmp = gl_adb->cbx;
	while (tmp != *cbx && tmp->next != *cbx) tmp = tmp->next;
	if (tmp && tmp != *cbx) *cbx = tmp;
	else while (*cbx && (*cbx)->next) *cbx = (*cbx)->next;
      }
    } else {
      if (key == KEYCODE_DOWN) {
	*sfo = (*sfo)->next;
	if (!*sfo) *sfo = gl_adb->sfo;
      } else {
	STRINFO_pt tmp = gl_adb->sfo;
	while (tmp != *sfo && tmp->next != *sfo) tmp = tmp->next;
	if (tmp && tmp != *sfo) *sfo = tmp;
	else while (*sfo && (*sfo)->next) *sfo = (*sfo)->next;
      }
    }
    break;
  case TAB_KEY:
    if (*cbx) {
      if (gl_adb->sfo) {
	STRINFO_pt tmp = gl_adb->sfo;
	while (tmp) {
	  if ((*cbx)->loc.y == tmp->loc.y) {
	    *cbx = (CHECKBOX_pt)NULL;
	    *sfo = tmp;
	    break;
	  }
	  tmp = tmp->next;
	}
	if (*sfo == (STRINFO_pt)NULL && (*cbx)->next == (CHECKBOX_pt)NULL) {
	  *sfo = gl_adb->sfo;
	  *cbx = (CHECKBOX_pt)NULL;
	}
      }
      if (*cbx) return(DLG_NextField(KEYCODE_DOWN, cbx, sfo));
    } else {
      if (gl_adb->cbx) {
	CHECKBOX_pt tmp = gl_adb->cbx;
	while (tmp) {
	  if ((*sfo)->loc.y < tmp->loc.y) {
	    *sfo = (STRINFO_pt)NULL;
	    *cbx = tmp;
	    break;
	  }
	  tmp = tmp->next;
	}
	if (*cbx == (CHECKBOX_pt)NULL && (*sfo)->next == (STRINFO_pt)NULL) {
	  *cbx = gl_adb->cbx;
	  *sfo = (STRINFO_pt)NULL;
	}
      }
      if (*sfo) return(DLG_NextField(KEYCODE_DOWN, cbx, sfo));
    }
    break;
  default:
    return(FALSE);
  }
  return(TRUE);
}

int DLG_PoseDialog(void)
{
  int key, bts_hit;
  STRINFO_pt sfo;
  CHECKBOX_pt cbx;

  cbx = gl_adb->cur_cbx;

  if (!cbx && gl_adb->cur_sfo) sfo = gl_adb->cur_sfo;
  else {
    cbx = gl_adb->cbx;
    if (!cbx) sfo = gl_adb->sfo;
  }

  if (cbx || sfo) {
    if (gNovice) SCR_WriteStatusLine(MSG_HintDialog);
    SCR_CursorOn();
  }

  while (TRUE) {
    SCR_SetCursorPosn(0, 0);
    if (cbx) {
      sfo = (STRINFO_pt)NULL;
      key = DLG_GetCheckBox(cbx);
      if (DLG_NextField(key, &cbx, &sfo)) continue;
    } else if (sfo) {
      cbx = (CHECKBOX_pt)NULL;
      key = DLG_GetStrField(sfo);
      if (DLG_NextField(key, &cbx, &sfo)) continue;
    } else key = SCR_GetKey();

    if (key == KEYCODE_REDRAW) {
      DLG_Draw();
      continue;
    }
    bts_hit = DLG_DoKey(key);
    if (bts_hit) return bts_hit;
    SCR_Beep();
  }
}

void DLG_Close(DIALOG_pt save_adb)
{
  BOX *adb = &gl_adb->bounds;

  gl_adb = save_adb;
  if (!gl_adb) {
    WIN_HLSelection(TRUE);
    SCR_CursorOn();
    WIN_RefreshRegion(adb->loc.y, adb->loc.x, adb->dim.y, adb->dim.x);
    WIN_HomeCursor();
    WIN_DoSetupMenus();
  } else {
    DLG_Draw();
  }
}

void DLG_About(void)
{
#define NUM_ABOUT_LINES (sizeof(about_text) / sizeof(char *))
#define ABOUT_HEIGHT    (NUM_ABOUT_LINES + START_TEXTROW + 2)
#define ABOUT_WIDTH     (60)

  int i, len;
  char title[STR255LEN];
  DIALOG about_dialog;
  DIALOG_pt prev_dialog;
  STEXT  astx[NUM_ABOUT_LINES];

  bzero(&about_dialog, sizeof(DIALOG));

  strcpy(title, getvstr());
  strcat(title, MSG_ReturnCloses);

  about_dialog.bounds.dim.x = ABOUT_WIDTH;
  about_dialog.bounds.dim.y = ABOUT_HEIGHT;

  about_dialog.title = title;

  len = strlen(about_text[0]);
  for (i=0; i<NUM_ABOUT_LINES; i++) {
    DLG_InitStaticText(&astx[i], i, (ABOUT_WIDTH-len-START_TEXTCOL)/2, 1,
		       len, about_text[i], (i >= 6 ? NORMCI : RINVIS));
    astx[i].next = (i < NUM_ABOUT_LINES-1 ? &astx[i+1] : NULL);
  }

  about_dialog.stx = astx;

  prev_dialog = DLG_Setup(&about_dialog, 0, 0, 0);

  DLG_PoseDialog();

  DLG_Close(prev_dialog);
}

int DLG_Login(char *userid, char *imap_host, char *smtp_host)
{
#define LOGIN_HEIGHT    (NUM_LOGIN_FIELDS * 2 + START_TEXTROW + START_BUTTON)
#define LOGIN_WIDTH     (46)

  int i, bts_hit;
  DIALOG login_dialog;
  DIALOG_pt prev_dialog;
  BUTTON bts[2];
  STEXT  stx[NUM_LOGIN_FIELDS];
  STRINFO sfo[NUM_LOGIN_FIELDS];
  char *login_buffer[3];
  login_buffer[0] = userid;
  login_buffer[1] = imap_host;
  login_buffer[2] = smtp_host;

  bzero(&login_dialog, sizeof(DIALOG));

  login_dialog.title = standard_title;
  login_dialog.bounds.dim.x = LOGIN_WIDTH;
  login_dialog.bounds.dim.y = LOGIN_HEIGHT;

  for (i=0; i<NUM_LOGIN_FIELDS; i++) {
    DLG_InitStaticText(&stx[i], i*2, 0, 1, 26, login_prompts[i], DBOX_HILITE);
    DLG_InitStrInfo(&sfo[i], login_buffer[i],
		    LINE_WIDTH, i*2, 13, LOGIN_WIDTH, NORMAL);
    if (i < NUM_LOGIN_FIELDS-1) {
      stx[i].next = &stx[i+1];
      sfo[i].next = &sfo[i+1];
    }    
  }

  login_dialog.stx = &stx[0];
  login_dialog.bts = &bts[0];
  login_dialog.sfo = &sfo[0];

  prev_dialog = DLG_Setup(&login_dialog, 2,
			  (OK_BUTTON | CANCEL_BUTTON), OK_BUTTON);

  while (TRUE) {
    if (!userid[0]) login_dialog.cur_sfo = &sfo[0];
    else if (!imap_host[0]) login_dialog.cur_sfo = &sfo[2];
    else if (!smtp_host[0]) login_dialog.cur_sfo = &sfo[3];
    else login_dialog.cur_sfo = &sfo[0];
    bts_hit = DLG_PoseDialog();
    if (bts_hit == CANCEL_BUTTON ||
	(userid[0] && imap_host[0] && smtp_host[0])) break;
  }

  DLG_Close(prev_dialog);

  return((bts_hit == OK_BUTTON) ? TRUE : FALSE);
}

int DLG_GetPasswordFromUser(char *prompt,char *result)
{
#define GET_STR_HEIGHT  (2 + 2 + START_TEXTROW + START_BUTTON)
#define GET_STR_WIDTH   65

  int bts_hit;
  DIALOG adialog;
  STRINFO sfo;
  STEXT   stx;
  BUTTON  bts[2];
  DIALOG_pt prev_dialog;

  bzero(&adialog, sizeof(DIALOG));

  adialog.bounds.dim.x = GET_STR_WIDTH;
  adialog.bounds.dim.y = GET_STR_HEIGHT;
  adialog.title = standard_title;

  DLG_InitStaticText(&stx, 0, 0, 2, GET_STR_WIDTH-4, prompt, DBOX_HILITE);

  DLG_InitStrInfo(&sfo, result, STR255LEN, 2, 0, GET_STR_WIDTH, NORMAL);

  sfo.attr = INVISI;

  adialog.stx = &stx;
  adialog.sfo = &sfo;
  adialog.bts = bts;

  prev_dialog = DLG_Setup(&adialog, 2,
			  (OK_BUTTON | CANCEL_BUTTON), OK_BUTTON);

  bts_hit = DLG_PoseDialog();

  DLG_Close(prev_dialog);

  if (bts_hit == OK_BUTTON)
      return TRUE;
  else
      return FALSE;
}

int DLG_GetStringFromUser(char *prompt,char *result)
{
#define GET_STR_HEIGHT  (2 + 2 + START_TEXTROW + START_BUTTON)
#define GET_STR_WIDTH   65

  int bts_hit;
  DIALOG adialog;
  STRINFO sfo;
  STEXT   stx;
  BUTTON  bts[2];
  DIALOG_pt prev_dialog;

  bzero(&adialog, sizeof(DIALOG));

  adialog.bounds.dim.x = GET_STR_WIDTH;
  adialog.bounds.dim.y = GET_STR_HEIGHT;
  adialog.title = standard_title;

  DLG_InitStaticText(&stx, 0, 0, 2, GET_STR_WIDTH-4, prompt, DBOX_HILITE);

  DLG_InitStrInfo(&sfo, result, STR255LEN, 2, 0, GET_STR_WIDTH, NORMAL);

  adialog.stx = &stx;
  adialog.sfo = &sfo;
  adialog.bts = bts;

  prev_dialog = DLG_Setup(&adialog, 2,
			  (OK_BUTTON | CANCEL_BUTTON), OK_BUTTON);

  bts_hit = DLG_PoseDialog();

  DLG_Close(prev_dialog);

  return(bts_hit == OK_BUTTON ? TRUE : FALSE);
}

int DLG_GetBooleanFromUser(char *prompt,int default_item)
{
#define GET_BOOL_HEIGHT (2 + 2 + START_TEXTROW + START_BUTTON)
#define GET_BOOL_WIDTH  65

  int bts_hit;
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  BUTTON bts[2];
  STEXT stx;
  char bigbuf[512];
  
  DLG_FormatText(prompt, &bigbuf[0], GET_BOOL_WIDTH - 4);

  bzero(&adialog, sizeof(DIALOG));

  adialog.bounds.dim.x = GET_BOOL_WIDTH;
  adialog.bounds.dim.y = GET_BOOL_HEIGHT;
  adialog.title = "Press Return for Default, %s Otherwise";

  DLG_InitStaticText(&stx, 0, 0, 2, GET_BOOL_WIDTH-4, bigbuf, DBOX_HILITE);

  adialog.stx = &stx;
  adialog.bts =  bts;

  prev_dialog = DLG_Setup(&adialog, 2,
			  (YES_BUTTON | NO_BUTTON),
			  (default_item ? YES_BUTTON : NO_BUTTON));

  bts_hit = DLG_PoseDialog();

  DLG_Close(prev_dialog);

  return(bts_hit == YES_BUTTON ? TRUE : FALSE);
}

void DLG_Alert(int alert_type,char *alert_msg)
{
#define ATEXT_HEIGHT (8)
#define ATEXT_WIDTH  (56)
#define ALERT_HEIGHT (ATEXT_HEIGHT + START_TEXTROW + START_BUTTON)
#define ALERT_WIDTH  (60)

  char bigbuf[1024];
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  BUTTON bts;
  STEXT  stx;

  SCR_Beep();

  if (!alert_msg || !*alert_msg) return;

  bzero(&adialog, sizeof(DIALOG));

  adialog.bounds.dim.x = ALERT_WIDTH;
  adialog.bounds.dim.y = ALERT_HEIGHT;
  adialog.title = MSG_ReturnCloses;

  DLG_FormatText(alert_msg, &bigbuf[0], ATEXT_WIDTH);

  DLG_InitStaticText(&stx, 0, (ALERT_WIDTH - ATEXT_WIDTH - START_TEXTCOL) / 2,
		     ATEXT_HEIGHT, ATEXT_WIDTH, bigbuf, DBOX_HILITE);

  adialog.stx = &stx;
  adialog.bts = &bts;

  prev_dialog = DLG_Setup(&adialog, (alert_type >= ALERT_warning ? 1 : 0),
			  OK_BUTTON, OK_BUTTON);

  DLG_PoseDialog();

  DLG_Close(prev_dialog);
}

void DLG_Help(void)
{
  DLG_Alert(0, noconn_help);
}

void DLG_NewMail(char *msg)
{
  DLG_Alert(1, msg);
}

void DLG_IdleClose(char *msg)
{
  DLG_Alert(1, msg);
}

int DLG_ChooseFromList(char *prompt,int  *result)
{
#define CFL_HEIGHT      (8 + 2 + START_TEXTROW + START_BUTTON)
#define CFL_WIDTH       70

  int  bts_hit;
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  BUTTON bts[2];
  long list_size, list_start;

  *result = 0;

  bzero(&adialog, sizeof(DIALOG));
  adialog.bounds.dim.x = CFL_WIDTH;
  adialog.bounds.dim.y = CFL_HEIGHT;

  adialog.title = "Choose From List (Press RETURN for Ok, %s to Cancel)";

  adialog.bts = bts;
  prev_dialog = DLG_Setup(&adialog, 2,(OK_BUTTON | CANCEL_BUTTON), OK_BUTTON);

  if (LST_Create(SEND_RL, adialog.bounds.loc.y + START_TEXTCOL,
		 adialog.bounds.loc.x + 1,
		 CFL_HEIGHT - 6, CFL_WIDTH - 2, prompt) == 0) {
    GetListSize(SEND_RL, &list_size, &list_start);
    LST_SetListSize(SEND_RL, (long)list_size, (long)list_start,1);
    LST_Show(SEND_RL, TRUE);
    while (TRUE) {
      int key;
      SCR_SetCursorPosn(0, 0);
      key = SCR_GetKey();
      if (LST_DoListKey(SEND_RL, key)) continue;
      if ((bts_hit = DLG_DoKey(key)) != 0) break;
      SCR_Beep();
    }
    if (bts_hit == OK_BUTTON) *result = LST_GetTRSelection(SEND_RL) + 1;
    LST_Dispose(SEND_RL);
  } else
    CON_WriteLine("Not enough memory to create ChooseFromList dialog");

  DLG_Close(prev_dialog);

  return(*result ? TRUE : FALSE);
}



void DLG_LocalPrefs(void)
{
#define LPF_NUM_FIELDS  14
#define LPF_NUM_FIELDS_STR 7
#define LPF_FIRST_STR_FIELD 7
#define LPF_HEIGHT   ((LPF_NUM_FIELDS + 3) + START_TEXTROW + START_BUTTON)
#define LPF_WIDTH    70
#define LPF_HEADROW  0
#define LPF_TEXTROW  (LPF_HEADROW + 2)

  int i, bts_hit;
  char title[STR255LEN];
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  BUTTON bts[2];
  STEXT  hdr_stx, stx[LPF_NUM_FIELDS];
  STRINFO sfo[LPF_NUM_FIELDS_STR];
  CHECKBOX cbx[LPF_NUM_FIELDS];
  char lpf_bufs[LPF_NUM_FIELDS_STR][STR255LEN];

  bzero(&adialog, sizeof(DIALOG));

  strcpy(lpf_bufs[0], config_getstring("Email"     ,""));
  strcpy(lpf_bufs[1], config_getstring("TmpDir"    ,"/tmp/"));
  strcpy(lpf_bufs[2], config_getstring("ExternalEditor"    ,""));
  strcpy(lpf_bufs[3], config_getstring("SentFolder"    ,""));
  strcpy(lpf_bufs[4], config_getstring("DraftFolder",  "" ));
  strcpy(lpf_bufs[5], config_getstring("RealName","" ));
  strcpy(lpf_bufs[6], config_getstring("MmencodePath", "/usr/bin/mmencode" ));

  sprintf(title, "%s, %s to Save", standard_title,
	  SCR_GetKeyText(KEYCODE_LOCAL_SAVE));

  adialog.title = &title[0];
  adialog.bounds.dim.x = LPF_WIDTH;
  adialog.bounds.dim.y = LPF_HEIGHT;

  DLG_InitStaticText(&hdr_stx, LPF_HEADROW, (int)((LPF_WIDTH-18)/2), 1, 18,
		     "Local Preferences", DBOX_HILITE);

  hdr_stx.next = &stx[0];

  for (i=0; i<LPF_NUM_FIELDS; i++) {
    int row = i + LPF_TEXTROW;
    int len = strlen(lpf_prompts[i]);

    /* add check box if top ones */
    if (i < LPF_FIRST_STR_FIELD) {
	DLG_InitCheckBox(&cbx[i], row, 0, 0);
	if (i != LPF_FIRST_STR_FIELD -1) {
	    cbx[i].next = &cbx[i+1];
	}
    }
    DLG_InitStaticText(&stx[i], row, 4, 1, len, lpf_prompts[i], DBOX_HILITE);
    if (i < LPF_NUM_FIELDS - 1) {
      stx[i].next = &stx[i+1];
    }
    if (i >= LPF_FIRST_STR_FIELD) {
      int idx = i - LPF_FIRST_STR_FIELD;
      DLG_InitStrInfo(&sfo[idx], lpf_bufs[idx], STR255LEN, row,
		      stx[i].bounds.loc.x + 28,
		      LPF_WIDTH, NORMAL);
      if (i < LPF_NUM_FIELDS-1) sfo[idx].next = &sfo[idx+1];
    }
  }

  cbx[0].val = config_getswitch("LoginOnStartup",0);
  cbx[1].val = config_getswitch("CloseOnSend",0);
  cbx[2].val = config_getswitch("AutoExpunge",0);
  cbx[4].val = config_getswitch("RememberPasswords",1);
  cbx[5].val = config_getswitch("SmartAliasing",0);
  cbx[6].val = config_getswitch("SignMail",0);

  adialog.cbx =  cbx;
  adialog.bts =  bts;
  adialog.sfo =  sfo;
  adialog.stx = &hdr_stx;

  prev_dialog = DLG_Setup(&adialog, 2,
		   (CANCEL_BUTTON | SAVE_BUTTON), 
			  SAVE_BUTTON);

  bts_hit = DLG_PoseDialog();

  if (bts_hit != CANCEL_BUTTON) {

      /* xxx check boxes */
      config_setswitch("LoginOnStartup",cbx[0].val);
      config_setswitch("CloseOnSend",cbx[1].val);
      config_setswitch("AutoExpunge",cbx[2].val);
      config_setswitch("RememberPasswords",cbx[4].val);
      config_setswitch("SmartAliasing",cbx[5].val);
      config_setswitch("SignMail",cbx[6].val);

      config_setstring("Email",lpf_bufs[0]);
      config_setstring("TmpDir",lpf_bufs[1]);
      config_setstring("ExternalEditor",lpf_bufs[2]);
      config_setstring("SentFolder",lpf_bufs[3]);
      config_setstring("DraftFolder",lpf_bufs[4]);
      config_setstring("RealName",lpf_bufs[5]);
      config_setstring("MmencodePath", lpf_bufs[6] );

      config_save();
  }

  DLG_Close(prev_dialog);
}

#define RPF_signmail 0
#define RPF_purge    1
#define RPF_outgoing 2
#define RPF_printer  3

void DLG_RemotePrefs()
{
#define RPF_NUM_FIELDS 4
#define RPF_NUM_FIELDS_STR 4
#define RPF_FIRST_STR_FIELD 0
#define RPF_HEIGHT   ((RPF_NUM_FIELDS + 1) * 2 + START_TEXTROW + START_BUTTON)
#define RPF_WIDTH    70
#define RPF_HEADROW  0
#define RPF_TEXTROW  (RPF_HEADROW + 2)

  int i, bts_hit;
  char title[STR255LEN];
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  BUTTON bts[2];
  STEXT hdr_stx, stx[RPF_NUM_FIELDS];
  STRINFO sfo[RPF_NUM_FIELDS_STR];
  char rpf_bufs[RPF_NUM_FIELDS_STR][STR255LEN];
  
  bzero ( &adialog, sizeof ( DIALOG ) );
  bzero ( &stx, sizeof(STEXT)*RPF_NUM_FIELDS );
  bzero ( &hdr_stx, sizeof(STEXT) );

  strcpy ( rpf_bufs[0], config_getstring ( "ImapServer", "mail" ) );
  strcpy ( rpf_bufs[1], config_getstring ( "SmtpServer", "smtp" ) );
  strcpy ( rpf_bufs[2], config_getstring ( "Userid", "" ) );
  strcpy ( rpf_bufs[3], config_getstring ( "DefaultDomain", "" ) );

  sprintf ( title, "%s, %s to Save", 
	    standard_title,
	    SCR_GetKeyText ( KEYCODE_LOCAL_SAVE ) );
  
  adialog.title = &title[0];
  adialog.bounds.dim.x = RPF_WIDTH;
  adialog.bounds.dim.y = RPF_HEIGHT;
 
  DLG_InitStaticText(&hdr_stx, RPF_HEADROW, (int)((RPF_WIDTH-18)/2), 1, 18,
		     "Network Preferences", DBOX_HILITE);

  hdr_stx.next = &stx[0];
  
  for ( i = 0; i < RPF_NUM_FIELDS; i++ ) { 
    int row = i + RPF_TEXTROW;
    int len = strlen ( rpf_prompts[i] );

    DLG_InitStaticText ( &stx[i], row, 0, 1, len, rpf_prompts[i], DBOX_HILITE );

    if ( i < RPF_NUM_FIELDS - 1 ) { 
      stx[i].next = &stx[i+1];
    } 
    
    if ( i >= RPF_FIRST_STR_FIELD ) { 
      int idx = i - RPF_FIRST_STR_FIELD;
      DLG_InitStrInfo ( &sfo[idx], rpf_bufs[idx], STR255LEN, 
			row, stx[i].bounds.loc.x + 28,
			RPF_WIDTH, NORMAL );
      if ( i < RPF_NUM_FIELDS - 1 )
	sfo[idx].next = &sfo[idx+1];
      else
	sfo[idx].next = NULL;
    }
  }
  
  //adialog.cbx = cbx;
  adialog.bts = bts;
  adialog.sfo = sfo;
  adialog.stx = &hdr_stx;
  
  prev_dialog = DLG_Setup(&adialog, 2, 
			  (CANCEL_BUTTON | SAVE_BUTTON),
			  SAVE_BUTTON);

  bts_hit = DLG_PoseDialog();

  if ( bts_hit != CANCEL_BUTTON ) { 
    config_setstring ( "ImapServer",    rpf_bufs[0] );
    config_setstring ( "SmtpServer",    rpf_bufs[1] );
    config_setstring ( "Userid",        rpf_bufs[2] );
    config_setstring ( "DefaultDomain", rpf_bufs[3] );
    config_save();
  }

  DLG_Close(prev_dialog);
}

int DLG_GetAttachmentInfo ( char * filename,
			    char * attachmentNumber )
{
#define GAI_NUM_FIELDS 2
#define GAI_HEIGHT ((GAI_NUM_FIELDS + 3) + START_TEXTROW + START_BUTTON)
#define GAI_WIDTH    60
#define GAI_HEADROW  0
#define GAI_TEXTROW  (GAI_HEADROW + 2)

  int i, bts_hit;
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  BUTTON bts[2];
  STEXT stx[GAI_NUM_FIELDS];
  STRINFO sfo[GAI_NUM_FIELDS];
  static char * prompts[GAI_NUM_FIELDS] = {
    "Filename:",
    "Attachment number:"
  };

  char * bufs[GAI_NUM_FIELDS];

  bufs[0] = filename;
  bufs[1] = attachmentNumber;

  bzero(&adialog, sizeof(DIALOG));

  adialog.bounds.dim.x = GAI_WIDTH;
  adialog.bounds.dim.y = GAI_HEIGHT;
  adialog.title = standard_title;

  adialog.bts = bts;

  for ( i = 0; i < GAI_NUM_FIELDS; i++ ) { 
    
    DLG_InitStaticText ( &stx[i], 
			 i + GAI_TEXTROW,
			 0, 1, 
			 strlen ( prompts[i] ),
			 prompts[i],
			 DBOX_HILITE );
    
    if ( i < GAI_NUM_FIELDS - 1 ) { 
      stx[i].next = &stx[i+1];
    }

    DLG_InitStrInfo ( &sfo[i], 
		      bufs[i],
		      STR255LEN,
		      i + GAI_TEXTROW,
		      stx[i].bounds.loc.x + 28,
		      GAI_WIDTH,
		      NORMAL );
    if ( i < GAI_NUM_FIELDS - 1 )
      sfo[i].next = &sfo[i+1];
    else
      sfo[i].next = NULL;
  }

  adialog.sfo = sfo;
  adialog.stx = &stx[0];
  
  
  prev_dialog = DLG_Setup(&adialog, 
			  2,
			  (OK_BUTTON | CANCEL_BUTTON), 
			  OK_BUTTON);

  bts_hit = DLG_PoseDialog();

  DLG_Close(prev_dialog);

  return((bts_hit == OK_BUTTON) ? TRUE : FALSE);
}



char *pwd_prompts[3] = {
  "Please Enter Your Current Password: ",
  "Please Enter Your New Password: ",
  "Please Re-Enter Your New Password: "
};

int DLG_ChangePassword(char *p1,char *p2,char *p3)
{
#define CPWD_HEIGHT     (2 + START_TEXTROW + START_BUTTON)
#define CPWD_WIDTH      60
#define NUM_PW_FIELDS   3

  int i, bts_hit;
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  BUTTON bts[2];
  STEXT stx;
  STRINFO sfo;
  char *pw_bufs[3];
  
  pw_bufs[0] = p1;
  pw_bufs[1] = p2;
  pw_bufs[2] = p3;
  for (i=0; i<NUM_PW_FIELDS; i++) *(pw_bufs[i]) = 0;

  bzero(&adialog, sizeof(DIALOG));

  adialog.bounds.dim.x = CPWD_WIDTH;
  adialog.bounds.dim.y = CPWD_HEIGHT;
  adialog.title = standard_title;

  adialog.bts = bts;
  prev_dialog = DLG_Setup(&adialog, 2, (OK_BUTTON | CANCEL_BUTTON), OK_BUTTON);

  for (i=0; i<NUM_PW_FIELDS; i++) {
    int len = strlen(pwd_prompts[i]);
    DLG_InitStaticText(&stx, 0, 0, 1, len, pwd_prompts[i], DBOX_HILITE);
    DLG_InitStrInfo(&sfo, pw_bufs[i], LINE_WIDTH, 0, len, CPWD_WIDTH,INVISI);
    adialog.stx = &stx;
    adialog.sfo = &sfo;
    DLG_DrawContent();
    bts_hit = DLG_PoseDialog();
    if (bts_hit == CANCEL_BUTTON) break;
    SCR_Beep();
  }

  DLG_Close(prev_dialog);

  return((bts_hit == OK_BUTTON) ? TRUE : FALSE);
}

int DLG_MailForwarding(void)
{
#define MFD_HEIGHT     (START_TEXTROW + START_BUTTON + 2)
#define MFD_WIDTH      60

  int bts_hit;
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  STEXT stx;
  BUTTON bts[4];

  bzero(&adialog, sizeof(DIALOG));

  adialog.bounds.dim.x = MFD_WIDTH;
  adialog.bounds.dim.y = MFD_HEIGHT;
  adialog.title = "Return(Query), C(Clear), S(Set), %s(Cancel)";

  adialog.bts = bts;
  adialog.stx = &stx;

  DLG_InitStaticText(&stx, 0, (MFD_WIDTH-16)/2, 1, 15,
		     "Mail Forwarding", DBOX_HILITE);

  prev_dialog = DLG_Setup(&adialog, 4,
		   (QUERY_BUTTON | CANCEL_BUTTON | SET_BUTTON | CLEAR_BUTTON),
		   QUERY_BUTTON);

  bts_hit = DLG_PoseDialog();

  DLG_Close(prev_dialog);

  if (bts_hit == CLEAR_BUTTON) bts_hit = MFD_CLEAR;
  else if (bts_hit == QUERY_BUTTON) bts_hit = MFD_QUERY;
  else if (bts_hit == SET_BUTTON) bts_hit = MFD_SET;
  else bts_hit = 0;

  return(bts_hit);
}

char *fax_prompts[] = {
  "Name of Recipient:",
  "Address:",
  "",
  "",
  "Fax Phone:",
  "Voice Phone:",
  "Charge Account:",
  "Time to Send:",
  "Subject:"
};

#define FAX_NUM_FIELDS (sizeof(fax_prompts) / sizeof(char *))
#define FAX_STEXT_LEN 20
#define FAX_WIDTH     60
#define FAX_TEXTROW   2
#define FAX_HEIGHT    (FAX_NUM_FIELDS+START_TEXTROW+START_BUTTON+FAX_TEXTROW+1)

FAX_HEADERS_pt g_FaxHeaders = (FAX_HEADERS_pt)NULL;

int DLG_FaxHeaders(int *use_more_options)
{
  int i, bts_hit;
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  STEXT hdr_stx, stx[FAX_NUM_FIELDS];
  STRINFO sfo[FAX_NUM_FIELDS];
  BUTTON  bts[3];
  char title[STR255LEN];

  *use_more_options = FALSE;

  if (g_FaxHeaders == (FAX_HEADERS_pt)NULL) {
    g_FaxHeaders = (FAX_HEADERS_pt)malloc(sizeof(FAX_HEADERS));
    if (g_FaxHeaders == (FAX_HEADERS_pt)NULL) {
      DLG_Alert(ALERT_warning, MSG_NotEnoughMemory);
      return(FALSE);
    }
    bzero(g_FaxHeaders, sizeof(FAX_HEADERS));
    strcpy(g_FaxHeaders->fax_destinfo[FAX_INFO_TIME], "Now");
    strcpy(g_FaxHeaders->fax_options[FAX_OPT_RETRY], "1");
    for (i=FAX_OPT_RES; i<=FAX_OPT_TRAIL; i++)
      g_FaxHeaders->fax_bools[i] = TRUE;
  }

  bzero(&adialog, sizeof(DIALOG));

  sprintf(title, "%s, %s for More",
	  standard_title, SCR_GetKeyText(KEYCODE_ALT_M));

  adialog.title = title;
  adialog.bounds.dim.x = FAX_WIDTH;
  adialog.bounds.dim.y = FAX_HEIGHT;

  DLG_InitStaticText(&hdr_stx, 0, (FAX_WIDTH - 26) / 2, 1, 27,
		     "Fax Destination Information", DBOX_HILITE);

  hdr_stx.next = &stx[0];

  for (i=0; i<FAX_NUM_FIELDS; i++) {
    int row = i + FAX_TEXTROW;
    DLG_InitStaticText(&stx[i], row, 0, 1,
		       strlen(fax_prompts[i]), fax_prompts[i], DBOX_HILITE);
    DLG_InitStrInfo(&sfo[i], g_FaxHeaders->fax_destinfo[i], STR255LEN,
		    row, FAX_STEXT_LEN, FAX_WIDTH, NORMAL);
    if (i < FAX_NUM_FIELDS - 1) {
      stx[i].next = &stx[i+1];
      sfo[i].next = &sfo[i+1];
    }
  }

  adialog.sfo = &sfo[0];
  adialog.bts = &bts[0];
  adialog.stx = &hdr_stx;

  prev_dialog = DLG_Setup(&adialog, 3,
			  (OK_BUTTON | CANCEL_BUTTON | MORE_BUTTON),OK_BUTTON);

  while (TRUE) {
    bts_hit = DLG_PoseDialog();
    if (bts_hit == MORE_BUTTON) {
      bts_hit = DLG_FaxHeadersOptions();
      if (bts_hit == OK_BUTTON) *use_more_options = TRUE;
    }
    else if (bts_hit == CANCEL_BUTTON) break;
    else {
      if (!g_FaxHeaders->fax_destinfo[FAX_INFO_PHONE][0])
	adialog.cur_sfo = &sfo[FAX_INFO_PHONE];
      else if (!g_FaxHeaders->fax_destinfo[FAX_INFO_TIME][0])
	adialog.cur_sfo = &sfo[FAX_INFO_TIME];
      else if (!g_FaxHeaders->fax_destinfo[FAX_INFO_SUBJECT][0])
	adialog.cur_sfo = &sfo[FAX_INFO_SUBJECT];
      else break;
      SCR_Beep();
    }
  }

  DLG_Close(prev_dialog);

  return(bts_hit == OK_BUTTON ? TRUE : FALSE);
}

char *fax_prompts2[] = {
  "Coverpage path:",
  "Bodyfile Path:",
  "Maximum Retries:",
  "Acknowledge",
  "Fine Resolution",
  "Confidential",
  "Trailing Coverpage"
};

#define FAX_NUM_FIELDS2 (sizeof(fax_prompts2) / sizeof(char *))
#define FAX_STEXT_LEN2 20
#define FAX_WIDTH2   (FAX_WIDTH - 4)
#define FAX_HEIGHT2  (FAX_HEIGHT - 2)

int DLG_FaxHeadersOptions(void)
{
  int i, bts_hit;
  DIALOG adialog;
  DIALOG_pt prev_dialog;
  STEXT hdr_stx, stx[FAX_NUM_FIELDS2];
  STRINFO sfo[3];
  CHECKBOX cbx[4];
  BUTTON  bts[2];

  bzero(&adialog, sizeof(DIALOG));

  adialog.title = standard_title;
  adialog.bounds.dim.x = FAX_WIDTH2;
  adialog.bounds.dim.y = FAX_HEIGHT2;

  DLG_InitStaticText(&hdr_stx, 0, (FAX_WIDTH2 - 26) / 2, 1, 22,
		     "Additional Fax Options", DBOX_HILITE);

  hdr_stx.next = &stx[0];

  for (i=0; i<3; i++) {
    int row = i + FAX_TEXTROW;
    DLG_InitStaticText(&stx[i], row, 4, 1,
		       strlen(fax_prompts2[i]), fax_prompts2[i], DBOX_HILITE);
    stx[i].next = &stx[i+1];
    DLG_InitStrInfo(&sfo[i], g_FaxHeaders->fax_options[i], STR255LEN,
		    row, FAX_STEXT_LEN2 + 4, FAX_WIDTH2, NORMAL);
    if (i < 2) sfo[i].next = &sfo[i+1];
  }

  for (i=3; i<FAX_NUM_FIELDS2; i++) {
    int idx = i - 3;
    int row = i + FAX_TEXTROW + 1;
    int col = 0;
    if (i % 2 == 0) {
      row--; col += 20;
    }
    DLG_InitStaticText(&stx[i], row, col+4, 1,
		       strlen(fax_prompts2[i]), fax_prompts2[i], DBOX_HILITE);
    DLG_InitCheckBox(&cbx[idx], row, col, (int)g_FaxHeaders->fax_bools[idx]);
    if (i < FAX_NUM_FIELDS2 - 1) {
      cbx[idx].next = &cbx[idx+1];
      stx[i].next = &stx[i+1];
    }
  }

  adialog.cbx = &cbx[0];
  adialog.sfo = &sfo[0];
  adialog.bts = &bts[0];
  adialog.stx = &hdr_stx;

  prev_dialog = DLG_Setup(&adialog, 2, (OK_BUTTON | CANCEL_BUTTON), OK_BUTTON);

  bts_hit = DLG_PoseDialog();

  DLG_Close(prev_dialog);

  if (bts_hit == CANCEL_BUTTON) {
    bzero(g_FaxHeaders->fax_options, sizeof(g_FaxHeaders->fax_options));
    strcpy(g_FaxHeaders->fax_options[FAX_OPT_RETRY], "1");
    g_FaxHeaders->fax_bools[0] = FALSE;
    for (i=FAX_OPT_CONF; i<=FAX_OPT_TRAIL; i++)
      g_FaxHeaders->fax_bools[i] = TRUE;
  } else {
    for (i=FAX_OPT_ACK; i<=FAX_OPT_TRAIL; i++)
      g_FaxHeaders->fax_bools[i] = cbx[i].val;
  }

  return(bts_hit);
}


void DLG_InitStrInfo(STRINFO *asi,char *buf,int bufsize, int row, int col, int w, int attr)
{
  bzero(asi, sizeof(STRINFO));

  col += START_TEXTCOL;

  asi->buf = buf;
  asi->bufsize = imax(0, bufsize);
  asi->scrsize = imax(0, (w - col - 4)); /* assume rest of the line */
  asi->loc.y = row + START_TEXTROW;
  asi->loc.x = col + 1;
  asi->attr = attr;
}

void DLG_InitStaticText(STEXT_pt stx,int r,int c,int h,int w,char *text,int attr)
{
  bzero(stx, sizeof(STEXT));

  stx->bounds.loc.y = r + START_TEXTROW;
  stx->bounds.loc.x = c + START_TEXTCOL;
  stx->bounds.dim.x = w;
  stx->bounds.dim.y = h;
  stx->attr = attr;
  stx->text = text;
}

int DLG_CheckButton(int key,int check_first_char)
{
  BUTTON_pt bts = gl_adb->bts;

  if (key == KEYCODE_FAKERETURN) key = KEYCODE_RETURN;

  while (bts) {
    if (bts->key == key || bts->key == SCR_UpCaseRepresentation(key)) break;
    bts = bts->next;
  }

  if (!bts && check_first_char) {
    bts = gl_adb->bts;
    while (bts) {
      int upkey = SCR_UpCaseRepresentation(key);
      if (*bts->text == key || *bts->text == upkey) break;
      bts = bts->next;
    }
  }

  if (bts) DLG_FlashButton(bts);

  return(bts ? bts->mask : 0);
}

void DLG_SetupButtons(POINT *dim,BUTTON_pt bts,
		      int num_bts, int bts_mask, int default_mask)
{
#define BUTTON_WIDTH   9
#define BUTTON_OFFSET  3

  static char *bts_names[] = {"Cancel", "Save...", "Ok", "No", "Yes",
                              "Set...", "Clear", "Query", "More..."};
  static int   bts_keys[]  = {CANCEL_KEY, KEYCODE_LOCAL_SAVE, KEYCODE_RETURN,
			      CANCEL_KEY, KEYCODE_RETURN, 0, 0, KEYCODE_RETURN, KEYCODE_ALT_M};
  int i, j, w, key, spacing;
  Boolean seen_default = FALSE;

  spacing = imax(0, (dim->x - (BUTTON_WIDTH * num_bts))) / (num_bts + 1);

  for (j=0, w=spacing, i=0x01; i<=MORE_BUTTON; j++, i=i<<1) {
    if (bts_mask & i) {
      if (default_mask == i) {
	key = KEYCODE_RETURN;
	seen_default = TRUE;
      } else
	key = (seen_default && bts_keys[j]==KEYCODE_RETURN) ? CANCEL_KEY : bts_keys[j];
      DLG_MakeButton(bts, dim->y - BUTTON_OFFSET, w, BUTTON_WIDTH,
		     bts_names[j], key, i, (default_mask == i));
      w += (BUTTON_WIDTH + spacing);
      bts_mask &= ~i;
      if (bts_mask) {
	bts->next = bts + 1;
	bts++;
	num_bts--;
      }
      else break;
    }
  }
}

void DLG_MakeButton(BUTTON_pt bts,int r, int c, int w,		    
		    char *name,int key, int mask, int is_default)
{
#define BUTTON_HEIGHT  1

  bzero(bts, sizeof(BUTTON));

  bts->bounds.loc.y = r;
  bts->bounds.loc.x = c;
  bts->bounds.dim.x = w;
  bts->bounds.dim.y = BUTTON_HEIGHT;
  bts->text = name;
  bts->is_default = is_default;
  bts->key = key;
  bts->mask = mask;
}

void DLG_DrawButton(BUTTON_pt abutton,int attr)
{
  char bname[LINE_WIDTH];
  int  len = strlen(abutton->text);
  int  w = abutton->bounds.dim.x;

  memset(bname, SPC_KEY, w);
  bcopy(abutton->text, &bname[(w-len)/2], len);
  if (abutton->is_default) {
    bname[0]   = '[';
    bname[w-1] = ']';
  }
  bname[w] = 0;

  DLG_ShowText(&abutton->bounds, bname, w, attr);
}

void DLG_FlashButton(BUTTON *abutton)
{
  DLG_DrawButton(abutton, NORMAL);
  DLG_DrawButton(abutton, DBOX_BUTTON);
  DLG_DrawButton(abutton, NORMAL);
  DLG_DrawButton(abutton, DBOX_BUTTON);
}

void DLG_InitCheckBox(CHECKBOX_pt cbx,int r,int c,int val)
{
  bzero(cbx, sizeof(CHECKBOX));

  cbx->loc.x = c + START_TEXTCOL;
  cbx->loc.y = r + START_TEXTROW;
  cbx->val = val;
}

int DLG_GetCheckBox(CHECKBOX_pt cbx)
{
  int key;
  while(TRUE) {
    DLG_SetCursor(cbx->loc.y, cbx->loc.x+1);
    key = SCR_GetKey();
    if (key == 'x' || key == 'X') {
      cbx->val = !cbx->val;
      DLG_DrawCheckBox(cbx);
    } else
      break;
  }
  return(key);
}

void DLG_DrawCheckBox(CHECKBOX_pt cbx)
{
#ifdef IBMPC
#define CHECKED_CHAR 0xFB
#else
#define CHECKED_CHAR 'X'
#endif
#define CBX_LENGTH    3

  char buf[CBX_LENGTH+1];

  buf[0] = '[';
  buf[2] = ']';
  buf[1] = ((cbx->val) ? CHECKED_CHAR : SPC_KEY);
  DLG_Write(buf, cbx->loc.y, cbx->loc.x, CBX_LENGTH, DBOX_BUTTON);
}

int DLG_GetStrField(STRINFO *asi)
{
  asi->datalen = strlen(asi->buf);
  DLG_DrawStrField(asi);

  DLG_SetCursor(asi->loc.y, asi->loc.x);
  SCR_UpdateScreen();

  asi->curpos = 0;
  asi->chrpos = 0;

  while(TRUE) {
    int ch = SCR_GetKey();
    if (DLG_ProcessKey(asi, ch) == FALSE) {
      asi->buf[asi->datalen] = 0;
      StripWhiteEnds(asi->buf);
      return(ch);
    }
    SCR_UpdateScreen();
  }
}

void DLG_DrawStrField(STRINFO_pt sfo)
{
  BOX bounds;

  bounds.loc.x = sfo->loc.x - 1;
  bounds.loc.y = sfo->loc.y;
  bounds.dim.x = bounds.dim.y = 1;
  DLG_ShowText(&bounds, "<", 1, DBOX_PAINT);
  bounds.loc.x += (sfo->scrsize + 2);
  DLG_ShowText(&bounds, ">", 1, DBOX_PAINT);
  bounds.loc.x = sfo->loc.x;
  bounds.dim.x = sfo->scrsize+1;
  DLG_ShowText(&bounds, sfo->buf, sfo->scrsize+1, sfo->attr);
}

int DLG_ProcessKey(STRINFO *asi,int key)
{
  switch (key) {
  case KEYCODE_RIGHT:
    if (asi->curpos >= asi->scrsize) {
      if (DLG_ShiftData(asi, 1)) return(TRUE);
    } else if (asi->chrpos+asi->curpos < asi->datalen) {
      DLG_SetCursor(asi->loc.y, ++(asi->curpos) + asi->loc.x);
      return(TRUE);
    }
    break;
  case KEYCODE_LEFT:
    if (asi->curpos <= 0) {
      if (DLG_ShiftData(asi, -1)) return(TRUE);
    } else {
      DLG_SetCursor(asi->loc.y, --(asi->curpos) + asi->loc.x);
      return(TRUE);
    }
    break;
  case KEYCODE_DELETE:
    if (asi->curpos + asi->chrpos >= asi->datalen) break;
  case KEYCODE_DEL:
  case KEYCODE_BACKSPACE:
    if (asi->datalen == 0) break;
    if (key != KEYCODE_DELETE) {
      if (asi->curpos <= 0 && asi->chrpos <= 0)
	break;
      DLG_ProcessKey(asi, KEYCODE_LEFT);
    }
    {
      int pos = asi->curpos + asi->chrpos;
      bcopy(&asi->buf[pos + 1], &asi->buf[pos], imax(0, asi->datalen - pos));
    }
    asi->datalen = imax(0, asi->datalen-1);
    asi->buf[asi->datalen] = 0;
    DLG_Write(&asi->buf[asi->chrpos],
	      asi->loc.y, asi->loc.x, asi->scrsize, asi->attr);
    DLG_SetCursor(asi->loc.y, asi->loc.x+asi->curpos);
    return(TRUE);
  case KEYCODE_HOME:
    asi->curpos = 0;
    DLG_ShiftData(asi, -(asi->datalen));
    return(TRUE);
  case KEYCODE_END:
    asi->curpos = (asi->scrsize <= asi->datalen ? asi->scrsize : asi->datalen);
    DLG_ShiftData(asi, asi->datalen);
    return(TRUE);
  case KEYCODE_CTRL_K:
    asi->buf[asi->chrpos+asi->curpos] = '\0';
    asi->datalen = strlen(asi->buf);
    DLG_Write(NIL, asi->loc.y, asi->loc.x + asi->curpos, 
	      asi->scrsize - asi->curpos, NORMAL);
    DLG_SetCursor(asi->loc.y, asi->loc.x + asi->curpos);
    return(TRUE);
  default:
    if (DLG_ValidKey(key)) {
      if (DLG_InsertChar(asi, key)) return(TRUE);
    } else
      return(FALSE);
  }
  SCR_Beep();
  return(TRUE);
}

int DLG_ShiftData(STRINFO *asi,int delta)
{
  char *str;

  str = asi->buf + asi->chrpos + delta;
  if (str < asi->buf) {
    str = asi->buf;
    asi->chrpos = 0;
  }
  else if (str >= asi->buf+asi->datalen) {
    asi->chrpos = imax(0, (asi->datalen - asi->scrsize));
    str = asi->buf + asi->chrpos;
  } else {
    if (asi->datalen - asi->chrpos < asi->scrsize) return(FALSE);
    asi->chrpos += delta;
  }

  DLG_Write(str, asi->loc.y, asi->loc.x, asi->scrsize, asi->attr);

  DLG_SetCursor(asi->loc.y, asi->loc.x+asi->curpos);

  return(TRUE);
}

int DLG_ValidKey(int key)
{
  return((isascii(key) && isprint(key)) ? TRUE : FALSE);
}

int DLG_InsertChar(STRINFO *asi,int key)
{
  int pos = asi->chrpos + asi->curpos;

  if ((pos+1) >= asi->bufsize) return(FALSE);

  bcopy(&asi->buf[pos], &asi->buf[pos+1], imax(0, asi->datalen - pos));
  asi->buf[pos] = (char)key;
  asi->datalen++;
  asi->buf[asi->datalen] = 0;

  if (asi->curpos >= asi->scrsize)
    DLG_ShiftData(asi, 1);
  else {
    DLG_Write(&asi->buf[asi->chrpos],
	      asi->loc.y, asi->loc.x+asi->chrpos, asi->scrsize, asi->attr);
    asi->curpos++;
  }
  DLG_SetCursor(asi->loc.y, asi->loc.x+asi->curpos);

  return(TRUE);
}

void DLG_Write(char *text,int r,int c,int w,int attr)
{
  r += gl_adb->bounds.loc.y;
  c += gl_adb->bounds.loc.x;

  SCR_Write(text, r, c, w, attr);
}

void DLG_ShowText(BOX *adb,char *text,int size, int attr)
{
  int r, c;

  r = adb->loc.y;
  c = adb->loc.x;

  while (size-- > 0) {
    if (c >= (adb->dim.x + adb->loc.x)) {
      c = adb->loc.x; r++;
    }
    if (r >= (adb->loc.y + adb->dim.y)) break;
    DLG_Write(text, r, c++, 1, 
	      (attr == RINVIS && text && *text == SPC_KEY)?DBOX_PAINT:attr);
    if (text && *text) text++;
  }
}

void DLG_SetCursor(int r,int c)
{
  r += gl_adb->bounds.loc.y;
  c += gl_adb->bounds.loc.x;
  SCR_SetCursorPosn(r, c);
}

void DLG_DrawBox(BOX *adb)
{
  SCR_Box(adb->dim.y, adb->dim.x,
	  adb->loc.y, adb->loc.x, DBOX_ATTR, 0L); /*, DOUBLE_LINE_BOX);*/
}

void DLG_Box(BOX *adb,char *title)
{
  int i, clear_row;

  DLG_DrawBox(adb);

  if (title && *title) {
    char buf[100];
    char ttl[STR255LEN];
    int tlen;
    SCR_Line(1, adb->dim.x-2, adb->loc.y+2, adb->loc.x+1,
	     DBOX_ATTR, 0L); /*, SINGLE_LINE_BOX);*/
    sprintf(ttl, title, SCR_GetKeyText(KEYCODE_ESCAPE));
    tlen = strlen(ttl);
    memset(buf, SPC_KEY, sizeof(buf));
    strcpy(buf+(imax(0, (adb->dim.x-tlen)/2)), ttl);
    DLG_Write(buf, 1, 1, adb->dim.x-2, RVIDEO);
    clear_row = 3;
  } else
    clear_row = 1;

  for(i=clear_row; i<adb->dim.y-1; i++)
    DLG_Write(NIL, i, 1, adb->dim.x-2, DBOX_PAINT);
}

void DLG_FormatText(char *text,char *buf,int w)
{
  if (text != NIL || w > 0) {
    int bytes, len = strlen(text);
    int c = 0;
    while (len > 0) {
      if (c >= w) c = 0;
      if (c == 0 && *text == ' ') {
	text++; len--;
	continue;
      }
      bytes = TXT_FormatText(text, len, w);
      if (bytes > 0 && text[bytes - 1] == '\n') bytes--;
      bcopy(text, buf, bytes);
      buf += bytes; text += bytes;
      len -= bytes; c += bytes; 
      while (c++ < w) *buf++ = ' ';
      if (*text == '\n') {
	text++; /* xxx was *text++ */
	len--;
      }
    }      
  }
  *buf = '\0';
}

void StripWhiteEnds(char *buf)
{
  char *savebuf = buf;
  if (buf) {
    int len;
    char *ptr;
    while (*buf == SPC_KEY) buf++; /* strip the front */
    len = strlen(buf);
    ptr = buf + (imax(0, len - 1));
    while (len && *ptr == SPC_KEY) {
      len--; *ptr-- = '\0';
    }
    strcpy(savebuf, buf);
  }
}
