/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <signal.h>

#include "mmsclie.h"
#include "mmclib.h"
#include "common.h"
#include "comkeys.h"
#include "menus.h"
#include "screen.h"
#include "list.h"
#include "and_keys.h"
#include "and_scr.h"
#include "dialog.h"
#include "console.h"
#include "windows.h"
#include "text.h"
#include "document.h"
#include "clinet.h"
#include "and_mach.h"
#include "smtp.h"
#include "ezconfig.h"
#include "attach.h"

extern const char *getprogname(void), *getformalprogname();

extern long bytes_transmitted;
extern unsigned char screen_rows, screen_cols;

int DoMenuCommand();

char *KeyMapFilename = "EzMail.key";

static char last_remote_str_buf[STR255LEN];
static int  last_remote_str_idx = -1;

static char last_local_str_buf[STR255LEN];
static int  last_local_str_idx = -1;

select_info gl_si;

char **gl_envp = 0;

Boolean gWeAreBusy = FALSE;
Boolean gShowFullHeaders = FALSE;
Boolean gFilerLongList = FALSE;
Boolean gCanSendFax = FALSE;
Boolean gBccMe    = FALSE;
Boolean gNovice   = TRUE;
Boolean gPuntRead = FALSE;
Boolean gCapturedShell = FALSE;

int g_BypassLogin = FALSE;

int g_MayWaitRefresh = FALSE;

int quit_in_progress = FALSE;
int net_is_connecting = FALSE;

extern void noconn_events();
extern void connecting_events();
extern void connected_events();
extern void quit_events();

#ifdef DESQview
extern int DESQview_version;
#endif

void (*wait_events)();

void setup_menus(void);

/* forwar declarations */
void change_forwarding(void);
void change_password(void);
int do_remote_command(int aCmdNumber);

int is_net_connecting(void)
{
  return(net_is_connecting);
}

void close_cleanup(void)
{
  int i;
  DOC_ReInitialize(TRUE);
  gShowFullHeaders = FALSE;
  WIN_DoSetupMenus();
  disconnect_all();
  for (i=FOLDER_RL; i<=DIRECTORY_RL; i++) {
    LST_Hide(i);
  }
}

void set_net_is_connecting(int aval)
{
  net_is_connecting = aval;
  if (net_is_connecting)
    wait_events = connecting_events;
  else
    wait_events = noconn_events;
}

static void connect_to_server(char *userid, char *imap, char *smtp)
{
    globals_t *g;
    char tmpstr[100];
    
    snprintf(tmpstr, sizeof(tmpstr)-1, "Connecting to %s...",imap);
    CON_SetBanner(tmpstr);
    CON_SetBusy(TRUE);
    
    set_net_is_connecting(TRUE);
    smtp_init(25);
    mn_clie_init(imap, userid, &g);

    CON_SetBusy(FALSE);
}

static void
connect_to_servers(void)
{
    char str[100];
    int i;
    char *imapdef = "mail";
    char *useriddef = (char *) config_getstring("Userid", getenv("USER"));

    for (i = 0; i < 5; i++) {
	const char *imap;
	const char *smtp;
	const char *userid;
	
	if (i > 0) 
	    snprintf(str, sizeof(str)-1,"ImapServer%d",i);
	else
	    snprintf(str, sizeof(str)-1,"ImapServer");
	imap = config_getstring(str,imapdef);

	if (i > 0) 
	    snprintf(str, sizeof(str)-1,"Userid%d",i);
	else
	    snprintf(str, sizeof(str)-1,"Userid");
	userid = config_getstring(str,useriddef);

	if (i > 0) 
	    snprintf(str, sizeof(str)-1,"SmtpServer%d",i);
	else
	    snprintf(str, sizeof(str)-1,"SmtpServer");
	smtp = config_getstring(str,"smtp");

	if (imap) {
	    char myuserid[LINE_LENGTH] = {0};
	    char myimap_host[LINE_LENGTH] = {0};
	    char mysmtp_host[LINE_LENGTH] = {0};

	    if (userid) {
		strncpy(myuserid, userid, sizeof(myuserid)-1);
	    }
	    if (imap) {
		strncpy(myimap_host, imap, sizeof(myimap_host)-1);
	    }
	    if (smtp) {
		strncpy(mysmtp_host, smtp, sizeof(mysmtp_host)-1);
	    }

	    /* JCS - Don't display the login if we have specified 
	     * to do it automatically on startup.  Why else do we 
	     * have this option? :-)
	     */
 	    if ((config_getswitch ( "LoginOnStartup", 0 )) || 
		(DLG_Login(myuserid, myimap_host, mysmtp_host))) {
		connect_to_server(myuserid, myimap_host, mysmtp_host);
	    }
	}

	imapdef = NULL;
    }

    do_all_showchanged();
    
    MNU_EnableItem(M_CONNECT, FALSE);
    MNU_EnableItem(M_DISCONNECT, TRUE);
    
    gFilerLongList = FALSE;   
}

static void disconnect_from_server(void)
{
    if (!CON_CanClose()) {
	CON_WriteLine(MSG_QuitAborted);
	return;
    }
        
    CON_SetBusy(TRUE);

    disconnect_all();
    
    set_net_dead(TRUE);
    
    /* xxx close all windows */
    /*    WIN_CloseFrontWindow();
    WIN_CloseFrontWindow();
    WIN_CloseFrontWindow();     */
    
    MNU_EnableItem(M_CONNECT, TRUE);
    MNU_EnableItem(M_DISCONNECT, FALSE);
    
    CON_SetBusy(FALSE);    
    
    if (quit_in_progress == TRUE) {
	SCR_RestoreVideoEnv();
	exit(0);
    }
}

void appl_ConnectHappened(void)
{

  set_net_is_connecting(FALSE);
  wait_events = connected_events;
  LST_Show(FOLDER_RL, CON_IsFrontWindow());
  WIN_DoSetupMenus();

  CON_SetBanner("Connected");

}

void SetBanner(int  list_no,char *banner)
{
  switch(list_no) {
  case CONSOLE_RL:
    CON_SetBanner(banner);
    break;
  case FOLDER_RL:
  case CAPTION_RL:
  case BODY_RL:
  case HELP_RL:
  case HELPBODY_RL:
  case DIRECTORY_RL:
    LST_SetBanner(list_no, (banner ? banner : nil_banners[list_no]));
  }
}

void ack_bytes(long apos)
{
  CON_SetBusy(FALSE);
  if (apos != bytes_transmitted) CON_SetBusy(TRUE);
}

int want_busy_command(void)
{
  if (gWeAreBusy) {
    SCR_Beep();
    SCR_WriteStatusLine(MSG_BusyMsg);
    return(FALSE);
  }
  return(TRUE);
}

int main(int argc,
	 char **argv, 
	 char **envp)
{

  gl_envp = envp; /* pointer to environment */

  signal(SIGPIPE, SIG_IGN);

  umask(066);  /* Set protection of temporary files */

  wait_events = noconn_events; /* start with no connection */

  SCR_SetupVideo();

  set_rl_window_size(RL_WINDOW_SIZE, (int)screen_rows);

  LST_init(RL_NUM_LISTS);
  WIN_init(setup_menus);

  MNU_initialize(DoMenuCommand);

  config_init();

  if (LST_CreateList(FOLDER_RL, nil_banners[FOLDER_RL]) ||
      LST_CreateList(CAPTION_RL, nil_banners[CAPTION_RL]) ||
      LST_CreateList(BODY_RL, nil_banners[BODY_RL]) ||
      LST_CreateList(HELP_RL, nil_banners[HELP_RL]) ||
      LST_CreateList(HELPBODY_RL, nil_banners[HELPBODY_RL]) ||
      LST_CreateList(DIRECTORY_RL, nil_banners[DIRECTORY_RL])) {
    SCR_RestoreVideoEnv();
    printf("Unable to create windows for lists\n");
    exit(-1);
  }

  CON_SetupConsole();
  CON_SetNoConnBanner();
  CON_Write((char*)getformalprogname());
  CON_WriteLine(" - A User Interface to the Andrew Message System.");

  DOC_ReInitialize(TRUE);

  read_args(argc, argv);

  MNU_SetupMenuBar();

  CON_Show();

  connect_to_servers();
  /*  if (config_getswitch("LoginOnStartup",0)) {
      connect_to_server();
  } else {
      SCR_Beep();
      DoMenuCommand(M_CONNECT);
      }*/
  

  mainloop();

  return 0;
}

void clear_checks(void)
{
  MNU_CheckItem(M_SHOW_FOLDERS,  FALSE);
  MNU_CheckItem(M_SHOW_CAPTIONS, FALSE);
  MNU_CheckItem(M_SHOW_MESSAGE,  FALSE);
  MNU_CheckItem(M_SHOW_CONSOLE,  FALSE);
  MNU_CheckItem(M_SHOW_FILER,    FALSE);
  MNU_CheckItem(M_SHOW_DOCUMENT, FALSE);
}

int DoMenuCommand(int menuid)
{
  char str[STR255LEN];

  str[0] = '\0';

  switch (menuid) {
  case M_REDRAW:
    WIN_RedrawScreen();
    break;
  case M_SUSPEND:
    SuspendProcess();
    break;
  case M_ABOUT:
    DLG_About();
    break;
  case M_MOVE:
    WIN_Move(GET_FrontWindow);
    break;
  case M_RESIZE:
    WIN_Resize(GET_FrontWindow);
    break;
  case M_ZOOM:
    WIN_Zoom(GET_FrontWindow);
    break;
  case M_CUT:
    WIN_DoKey(KEYCODE_CUT);
    break;
  case M_COPY:
  case M_EXCERPT:
    if (WIN_FrontWindowType == WTYPE_LIST) {
	
	LST_CopyText((menuid == M_EXCERPT),NULL); /* xxx NULL should be from */
    }
    else if (WIN_FrontWindowType == WTYPE_TEXT) TXT_CopyText();
    else SCR_Beep();
    break;
  case M_NEW:
    DOC_OpenNew();
    break;
  case M_CLOSE:
    WIN_CloseFrontWindow();
    break;
   case M_SHOW_FOLDERS:
    LST_Show(FOLDER_RL, TRUE);
    break;
  case M_SHOW_CAPTIONS:
    LST_Show(CAPTION_RL, TRUE);
    break;
  case M_SHOW_MESSAGE:
    LST_Show(BODY_RL, TRUE);
    break;
  case M_SHOW_FILER:
    if (!LST_ListShown(DIRECTORY_RL)) RL_do_num_command(DIRECTORY_RL,6,0L);
    LST_Show(DIRECTORY_RL, TRUE);
    break;
  case M_SHOW_CONSOLE:
    CON_Show();
    break;
  case M_SHOW_DOCUMENT:
    TXT_BringNextDocToFront();
    break;
  case M_EDITMSG:
    TXT_Edit();
    break;
  case M_EXTERNAL_EDIT:
    if (!gCapturedShell)
	TXT_ExtEdit();
    break;
  case M_OPEN_LOCAL:
    DOC_OpenLocalFile();
    break;
  case M_HELP:
    if (!is_net_dead()) {
      if (!LST_ListShown(HELP_RL)) RL_do_num_command(HELP_RL,0,0L);
      LST_Show(HELP_RL, TRUE);
    } else DLG_Help();
    break;
  case M_QUIT:      
    if (quit_in_progress) QuitCleared();
    else if (!CON_CanClose()) {
      CON_WriteLine(MSG_QuitAborted);
      break;
    }
    disconnect_all();
    quit_in_progress = TRUE;
  case M_DISCONNECT:
      disconnect_from_server();
    break;
  case M_CONNECT:
      connect_to_servers();
      break;
  case M_LOCAL_PREFS:
      DLG_LocalPrefs();
      break;
  case M_REMOTE_PREFS:
    DLG_RemotePrefs();
    break;
  case M_SIGN_THIS_MAIL:
    /*    SetRemotePrefsBoolean(BOOL_PREFS_BSIGN, gSignMail);*/
    break;
  case M_BCC_THIS_MAIL:
    gBccMe = !gBccMe;
    /*    SetRemotePrefsBoolean(BOOL_PREFS_BCOPY, gBccMe);*/
    MNU_CheckItem(M_BCC_THIS_MAIL, gBccMe);
    break;
  case M_SVC_PASSWORD:
    change_password();
    break;
  case M_SVC_FORWARD:
    change_forwarding();
    break;
  case M_APPEND:
  case M_APND_LOCAL:
    if (get_local_string(MSG_APPEND_LOCAL, str))
      TXT_DoMenuCommand(menuid, str);
    break;
  case M_BUG:
    DOC_BugReport();
    break;
  case M_KEYMAP:
    strcpy(str, KeyMapDirectory);
    strcat(str, KeyMapFilename);
    TXT_OpenBlurb(str, "EzMail Key Mapping");
    break;
  case M_PRINT_LOCAL:
    if (WIN_FrontWindowType == WTYPE_LIST)
      LST_SaveOrPrintList(NIL, TRUE);
    else
      TXT_PrintWindow();
    break;
  case M_SET_MARK:
    WIN_DoKey(KEYCODE_SET_MARK);
    break;
  case M_SEARCH:
      LST_search(CAPTION_RL);
      break;
  case M_SIGNMSG:
      return(TXT_DoMenuCommand(menuid, NULL));
  case M_QUOTA:
      do_all_showquota();
      break;
  case M_SAVE:
  case M_SAVE_LOCAL:
  case M_SAVE_REMOTE:
    if (LST_CaptionOrMessageInFront() && ( menuid != M_SAVE_LOCAL ) ) {
      menuid = M_FILE_IN_FILE;
    } else {
      int   save_it;
      if (WIN_FrontWindowType == WTYPE_TEXT)
	strcpy(str,WIN_FrontWindowTitle);
      if (menuid == M_SAVE_REMOTE)
	save_it = DLG_GetStringFromUser(remote_strings[9], str);
      else 
	save_it = get_local_string(MSG_SAVE_OR_OPEN_LOCAL, str);
      if (save_it) {
	if (WIN_FrontWindowType == WTYPE_LIST) {
	  return ( handle_save_all_to_file ( str ) );
	} else {
	  return(TXT_DoMenuCommand(menuid, str));
	}
      }
      break;
    }
    break;
  case M_SAVE_ATTACH:
    return handle_save_attachment();

  default:
      if (do_remote_command(menuid)) break;
      SCR_Beep();
      SCR_Beep();
  }

  return 0;
}

int do_remote_command(int aCmdNumber)
{
  int ListNumber, cmdIndex = 0, is_busy;
  long Selected, OpensDoc, strIdx, instances;
  char aStr[STR255LEN];

  aStr[0] = '\0';

  ListNumber = -1;
  Selected = 0;
  aStr[0] = 0;

  if (aCmdNumber == M_SEND || aCmdNumber == M_VERIFY)
  {
    return(TXT_DoMenuCommand(aCmdNumber, NIL));
  }

  if (aCmdNumber <= REMOTE_FILEINTO) {
    MNU_GetMenuName(aCmdNumber, aStr);
    if (aStr[0]) aCmdNumber = M_FILE_BYNAME;
  }

  if (aCmdNumber >= M_CREATE) { /* first possible remote option */
    cmdIndex = aCmdNumber % 1000;
    if (cmdIndex < 100) ListNumber = FOLDER_RL;
    else if (cmdIndex < 200) ListNumber = SEND_RL;
    else if (cmdIndex < 300) ListNumber = LST_PickCaptionOrMessageList();
    else ListNumber = DIRECTORY_RL;

    if (aCmdNumber != M_QUICK_EDIT) Selected = LST_GetTRSelection(ListNumber);
    cmdIndex = cmdIndex % 100;
  }

  if ((Selected != 0) || (ListNumber == SEND_RL))
    if ((aCmdNumber == M_POST) &&
	((ListNumber == CAPTION_RL) || (ListNumber == BODY_RL)))
      cmdIndex = M_C_POST - M_KILL;

  if (!GetRemoteCommandEntry(cmdIndex, ListNumber,
			    &strIdx, &instances, &OpensDoc, &is_busy))
    return(FALSE);

  if (aCmdNumber == M_FILE_BYNAME && aStr[0]) strIdx = -1;

  if (strIdx != -1) {
    char *prompt;
    prompt = remote_strings[strIdx];
    if (last_remote_str_idx != strIdx) last_remote_str_buf[0] = 0;
    last_remote_str_idx = strIdx;
    if (DLG_GetStringFromUser(prompt, last_remote_str_buf) == FALSE)
      return(TRUE);
    strcpy(aStr, last_remote_str_buf);
  }

  if (OpensDoc)
    DOC_ComposeMail(aCmdNumber,ListNumber,cmdIndex, Selected, aStr);
  else
    RL_do_command(ListNumber, cmdIndex, (long)Selected, aStr);

  if ((aCmdNumber == M_EXPOSE_MATCH) ||
      (aCmdNumber == M_EXPOSE_PERSONAL) ||
      (aCmdNumber == M_EXPOSE_ALL) ||
      (aCmdNumber == M_EXPOSE_SUBSCRIBED) ||
      (aCmdNumber == M_EXPOSE_CHANGED)) {
    RL_zero_list(FOLDER_RL);
    RL_zero_list(CAPTION_RL);
    RL_zero_list(BODY_RL);
    LST_Show(FOLDER_RL, TRUE);

  }
  else if (aCmdNumber == M_PUNT) {
    if (gPuntRead == FALSE) LST_Show(CAPTION_RL, TRUE);
  }
  else if (aCmdNumber == M_HEADERS) {
    gShowFullHeaders = (gShowFullHeaders == TRUE ? FALSE : TRUE);
    MNU_CheckItem(M_HEADERS, (int)gShowFullHeaders);
  } else if (aCmdNumber == M_AFS_LONGLIST) {
    gFilerLongList = (gFilerLongList == TRUE ? FALSE : TRUE);
    MNU_CheckItem(M_AFS_LONGLIST, (int)gFilerLongList);
  }
  if (aCmdNumber == M_APND_REMOTE) TXT_DoMenuCommand(M_APND_REMOTE, aStr);

  return(TRUE);
}

void setup_menus(void)
{
  int net_is_dead;

  net_is_dead = is_net_dead();
  if (net_is_dead)
    MNU_RevertToInitial();
  else
    MNU_EnableAll(TRUE);

  MNU_EnableItem(M_CONNECT, net_is_dead && !is_net_connecting());
  MNU_EnableItem(M_DISCONNECT, is_net_connecting() || !net_is_dead);
  MNU_EnableItem(M_SEND, FALSE);
  MNU_EnableItem(M_VERIFY, FALSE);
  MNU_EnableItem(M_SIGNMSG, FALSE);
  MNU_EnableItem(M_EDITMSG, FALSE);
  MNU_EnableItem(M_EXTERNAL_EDIT, FALSE);

  MNU_EnableItem(M_APPEND, FALSE);
  MNU_EnableItem(M_APND_LOCAL, FALSE);
  MNU_EnableItem(M_APND_REMOTE, FALSE);

  MNU_EnableItem(M_PUNT, FALSE);

  clear_checks();

  MNU_EnableItem(M_CUT, FALSE);
  MNU_EnableItem(M_COPY, FALSE);
  MNU_EnableItem(M_PASTE, FALSE);
  MNU_EnableItem(M_EXCERPT, FALSE);
  MNU_EnableItem(M_ADD_HEADERS, FALSE);
  MNU_EnableItem(M_ADD_FAXHDRS, FALSE);

  LST_SetupMenus(net_is_dead);
  CON_SetupMenus(net_is_dead);
  TXT_SetupMenus(net_is_dead);

  MNU_CheckItem(M_HEADERS, (int)gShowFullHeaders);
  MNU_CheckItem(M_BCC_THIS_MAIL, gBccMe);

  MNU_SetupMenuBar();
}

void get_boolean_from_user(char *prompt,int default_item, int lii)
{
  default_item = DLG_GetBooleanFromUser(prompt, default_item);
}

void get_string_from_user(char *prompt,int lii)
{
  char str[STR255LEN];
  str[0] = '\0';
  if (DLG_GetStringFromUser(prompt, str) == FALSE) str[0] = 0;
}

void choose_from_list(char *prompt,int default_item, int lii)
{
  int result;
  if (!DLG_ChooseFromList(prompt, &result)) result = default_item;
}

void change_password(void)
{
  char pw1[STR255LEN], pw2[STR255LEN], pw3[STR255LEN];

  if (DLG_ChangePassword(pw1, pw2, pw3)) {
    if ((strcmp(pw1, pw2) != 0) && (strcmp(pw2, pw3) == 0)) {
      char pbuf[STR255LEN];
      strcpy(pbuf, pw1); strcat(pbuf, "\015");
      strcat(pbuf, pw2);
      RL_do_command(SEND_RL, 11, -1L, pbuf);
    } else
      DLG_Alert(ALERT_warning, MSG_BadPwd);
  }
}

void change_forwarding(void)
{
  int val = DLG_MailForwarding();

  switch(val) {
  case MFD_QUERY:
    RL_do_num_command(SEND_RL, 10, (long)val);
    break;
  case MFD_CLEAR:
    RL_do_num_command(SEND_RL, 10, (long)val);
    break;
  case MFD_SET:
    {
      char aStr[STR255LEN];
      aStr[0] = '\0';
      if (DLG_GetStringFromUser("Please Enter Forwarding Address:", aStr)) {
	RL_do_command(SEND_RL, 10, (long)val, aStr);
      }
    }
  }
}

int get_local_string(int id,char *buf)
{
  int result;
  if (*buf == 0 && last_local_str_idx == id)
    strcpy(buf, last_local_str_buf);

  result = DLG_GetStringFromUser(local_strings[id], buf);
  strcpy(last_local_str_buf, buf);
  last_local_str_idx = id;
  return(result);
}

void HintUser(char *msg,int keytab[], int keytabsize)
{
  if (msg) {
    int i;
    char key_buf[MAX_HINT_KEYS][20], hint_buf[STR255LEN];
    for (i=0; i<MAX_HINT_KEYS; i++) key_buf[i][0] = 0;
    for (i=0; i<keytabsize; i++) {
      char *key_text = SCR_GetKeyText(keytab[i]);
      strcpy(key_buf[i], key_text);
    }
    sprintf(hint_buf, msg,
	    key_buf[0], key_buf[1], key_buf[2], key_buf[3], key_buf[4]);
    SCR_WriteStatusLine(hint_buf);
  } else
    SCR_ClearStatusLine();
}

void DisplayServerHint(char *hint_msg)
{
  extern int w_mode;

  if ((w_mode & (W_NOBORD_MODE|W_ZOOM_ALL)) || WindowZoomed(GET_FrontWindow)) {
    char hint_buf[STR255LEN], *cr_pos;
    strncpy(hint_buf, hint_msg, sizeof(hint_buf));
    hint_buf[sizeof(hint_buf)-1] = 0;
    if ((cr_pos = index(hint_buf, '\n')) != NIL) *cr_pos = 0;
    SCR_WriteStatusLine(hint_buf);
  }
}

void rl_error(char *str)
{
    CON_Write("rl error: ");
    CON_WriteLine(str);
}

int never_sel(void)
{
    printf("never sel\n");
    return -1;
}

int serv_conn_enable(void)
{
    printf("serv_conn_enable\n");
    return -1;
}

int fatal(char *msg, int code)
{
    printf("fatal: %s\n",msg);
    
    exit(0);
}
