/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * list management defines
 */

#ifndef LIST_H
#define LIST_H

#include "common.h"
#include "screen.h"

typedef struct list_struct {
  int  list_no;
  long size;
  long selection;
  long top;
  long mark_start, mark_end;
  WINLIST *win;
  char myban[MAX_BAN+1];
} LIST, *LIST_pt;

extern LIST_pt all_lists;

#define GET_LIST_pt(xxll) (all_lists+xxll)

#define SET_LIST_SIZE(xxll,xxsize) (GET_LIST_pt(xxll))->size = xxsize
#define SET_LIST_NUM(xxll) (GET_LIST_pt(xxll))->list_no = xxll
#define SET_LIST_SELECTION(xxll,xxpos) (GET_LIST_pt(xxll))->selection = xxpos
#define SET_LIST_TOP(xxll,xxtop) (GET_LIST_pt(xxll))->top = xxtop;

#define GET_LIST_SIZE(xxll) ((GET_LIST_pt(xxll))->size)
#define GET_LIST_SELECTION(xxll) ((GET_LIST_pt(xxll))->selection)
#define GET_LIST_TOP(xxll) ((GET_LIST_pt(xxll))->top)


void LST_init(int num);

extern long LST_GetTRSelection();

void LST_InitList(int  lno,long size, long first, long top,WINLIST *win);

int LST_DoListKey(int list_no,int key);

int LST_GetListidFromWindow(void);

int LST_PickCaptionOrMessageList(void);

int LST_CaptionOrMessageInFront(void);

void LST_SetBanner(int ll,char *aBan);

void LST_DisplayPosition(LIST_pt alist);

void LST_SetupMenus(int net_is_dead);

void LST_Hide(int ll);

int LST_SaveOrPrintList(char *fn,int print_it);

int LST_ListShown(int ll);

void LST_CopyText(int excerpt_it, const char *from);

int LST_CreateList(int list_no,char *banner);

void LST_Show(int ll, int activate);

int LST_Create(int ll, int r, int c, int h, int w,char *aban);

void LST_SetListSize(int  list_no,long list_size, long list_start,int resetpos);

void LST_Dispose(int ll);

void LST_SetSelection(int list_no,long item);

void LST_InvalidateItem(int  list_no,long item_no);

void LST_ReportSize(int lno);

void LST_ChangedSize(int lno);

void LST_SetRLTitle(int ln,char *title);

void LST_BringSelectionIntoView(int list_no);

void LST_search(int list_no);

#endif /* LIST_H */
