/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * menu operations module
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <strings.h>

#include "common.h"
#include "menus.h"
#include "screen.h"
#include "comkeys.h"
#include "and_keys.h"
#include "windows.h"
#include "and_scr.h"

#ifdef IBMPC
#define MENU_BORDER_ATTRIBUTE HILITE
#else /* IBMPC */
#define MENU_BORDER_ATTRIBUTE NORMAL
#endif /* IBMPC */

#ifdef IBMPC
#define H_OFF	BWHITE
#define H_ON	BLKGRY
#define H_DIM   DWHITE
#else /* IBMPC */
#define H_OFF   HILITE
#define H_ON    RVIDEO
#define H_DIM   NORMAL
#endif /* IBMPC */

#define INDENT 1
#define FLAG_COL (INDENT)
#define TEXT_COL (INDENT + 1)
#define TEXT_ROW (INDENT)

extern char *malloc(), *realloc();

static int (*MNU_DoMenuCommand)() = NULL_INT_FUNC_pt;

#ifdef IBMPC
char MNU_HARROW[] = {' ', 0x10, ' '};
char MNU_CHKMRK[] = {0xFB, '\0'};
char MNU_DISABLED_CHAR = '*';
#else
char MNU_HARROW[] = {'-', '>', ' '};
char MNU_CHKMRK[] = {'+', '\0'};
char MNU_DISABLED_CHAR = '*';
#endif

#define rem_save_or_print_str	"On Andrew..."
#define loc_save_or_print_str	"On Local Machine..."

#define rem_open_or_append_str  "Andrew File..."
#define loc_open_or_append_str  "Local File..."

#define COMMAND_KEY_WIDTH 6

#define MENU_PULLDOWN_KEY   '\r'

extern char *SCR_GetKeyText();

static int lastmenu_cmd = -1;
static int lastmenu_idx = -1;

static int menutree_idx = -1;
static int menuitem_idx = -1;
static int menuharc_idx = -1;
static int menubar_row  = 0;
static int menu_closekey = ' ';
static int menu_alt_closekey = KEYCODE_ESCAPE;
static Boolean menu_isopen = FALSE;

static Boolean menu_focus = FALSE;
static Boolean harc_focus = FALSE;

MENU Subscribe_menu[] = {
    {M_SUB_FULL,	    "Subscribe",FALSE,	NULL,	0,0,KEYCODE_SUBS_MENU,0},
    {M_SUB_NONE,	  "UnSubscribe",FALSE,	NULL,	0,0,0,0}
};
#define SUB_MENU_SIZE	(sizeof(Subscribe_menu) / sizeof(MENU))

MENULOC subscribe_menu_loc = {SUB_MENU_SIZE, 0,0,0,0,0};

MENU Respond_menu[] = {
    {M_REPLY_SENDER,	"Sender",   FALSE,  NULL,   0,0,KEYCODE_REPLY_MENU,0},
    {M_REPLY_READERS,	"Readers",  FALSE,  NULL,   0,0,0,0},
    {M_REPLY_ALL,	"All",	    FALSE,  NULL,   0,0,0,0},
    /*    {M_REPLY_RESEND,	"Resend",   FALSE,  NULL,   0,0,0,0},*/
    {M_REPLY_FORWARD,	"Forward",  FALSE,  NULL,   0,0,0,0},
    {M_REPLY_DRAFT,     "Draft Revive", FALSE, NULL, 0,0,0,0}
    /*{M_RESTORE,	        "Restore",  FALSE,  NULL,   0,0,0,0},*/
};
#define RESPOND_MENU_SIZE   (sizeof(Respond_menu) / sizeof(MENU))

MENULOC respond_menu_loc = {RESPOND_MENU_SIZE, 0,0,0,0,0};

MENU FileInto_menu[] = {
    {M_FILE_BYNAME,	"By Name...",  TRUE,  NULL,   0,0,KEYCODE_FMSG_MENU,0}
};
#define FILEINTO_MENU_SIZE  (sizeof(FileInto_menu) / sizeof(MENU))

MENU *alt_fileinto_menu = NULL;
MENULOC alt_fileinto_menu_loc = {0,0,0,0,0,0};
MENULOC fileinto_menu_loc = {FILEINTO_MENU_SIZE, 0,0,0,0,0};

MENU Open_menu[] = {
    /*    {M_OPEN_REMOTE,  rem_open_or_append_str,FALSE,NULL,0,0,0,0},*/
     {M_OPEN_LOCAL,   loc_open_or_append_str,TRUE, NULL,0,0,KEYCODE_LOCAL_OPEN,0}
};
#define OPEN_MENU_SIZE	(sizeof(Open_menu) / sizeof(MENU))

MENULOC open_menu_loc = {OPEN_MENU_SIZE, 0,0,0,0,0};

MENU Append_menu[] = {
    {M_APND_LOCAL,   loc_open_or_append_str,	TRUE,	NULL,	0,0,0,0},
    /*{M_APND_REMOTE,  rem_open_or_append_str,	FALSE,	NULL,	0,0,0,0}*/
};
#define APPEND_MENU_SIZE    (sizeof(Append_menu) / sizeof(MENU))

MENULOC append_menu_loc = {APPEND_MENU_SIZE, 0,0,0,0,0};

MENU Save_menu[] = {
    {M_SAVE_REMOTE,  rem_save_or_print_str,FALSE,NULL,0,0,0,0},
    {M_SAVE_LOCAL,   loc_save_or_print_str,TRUE, NULL,0,0,KEYCODE_LOCAL_SAVE,0},
    {M_SAVE_ATTACH, "Save Attachment", TRUE, NULL, 0, 0, KEYCODE_SAVE_ATTACH, 0 },
};
#define SAVE_MENU_SIZE	(sizeof(Save_menu) / sizeof(MENU))

MENULOC save_menu_loc = {SAVE_MENU_SIZE, 0,0,0,0,0};

MENU Expose_menu[] = {
    {M_EXPOSE_CHANGED,	"Changed",      FALSE,  NULL, 0,0,KEYCODE_EXPOSE_MENU,0},
    {M_EXPOSE_PERSONAL,	"Personal",	FALSE,  NULL, 0,0,0,0},
    {M_EXPOSE_SUBSCRIBED,"Subscribed",	FALSE,	NULL, 0,0,0,0},
    {M_EXPOSE_MATCH,	"Match...",	FALSE,  NULL, 0,0,0,0},
    {M_EXPOSE_ALL,	"All",		FALSE,  NULL, 0,0,0,0},
    {M_EXPOSE_BYNAME,    "By Name...",   FALSE,  NULL, 0,0,0,0}
};
#define EXPOSE_MENU_SIZE    (sizeof(Expose_menu) / sizeof(MENU))

MENULOC expose_menu_loc = {EXPOSE_MENU_SIZE, 0,0,0,0,0};

MENU Andrew_menu[] = {
    {M_FINGER,       "Finger...",          FALSE,NULL, 0,0,KEYCODE_UNIX_SVC,0},
    /*    {M_SVC_PASSWORD, "Change Password...", TRUE, NULL, 0,0,0,0}, */
    /* {M_SVC_FORWARD,  "Mail Forwarding...", TRUE, NULL, 0,0,0,0}, */
    /* {M_AFS_HELP,     "Andrew Help...",     TRUE, NULL, 0,0,0,0}, */
    /* {M_AFS_PRINTQ,   "Print Queue...",     TRUE, NULL, 0,0,0,0}, */
    /* {M_MOTD,	    "Show Motd",          FALSE,NULL, 0,0,0,0}*/
};
#define ANDREW_MENU_SIZE    (sizeof(Andrew_menu) / sizeof(MENU))

MENULOC andrew_menu_loc = {ANDREW_MENU_SIZE, 0,0,0,0,0};

MENU AFS_menu[] = {
    {M_AFS_HOME,     "Home",             TRUE,   NULL,   0,0,KEYCODE_AFS_SVC,0},
    {M_AFS_CAT,      "Show",             TRUE,   NULL,   0,0,0,0},
    {M_AFS_EDIT,     "Edit",             TRUE,   NULL,   0,0,0,0},
    {M_AFS_NEW,      "New...",           TRUE,   NULL,   0,0,0,0},
    {M_AFS_MV,       "Rename...",        TRUE,   NULL,   0,0,0,0},
    {M_AFS_RM,       "Delete",           TRUE,   NULL,   0,0,0,0},
    {M_AFS_CD,       "Change DIR...",    TRUE,   NULL,   0,0,0,0},
    {M_AFS_MKDIR,    "Make DIR...",      TRUE,   NULL,   0,0,0,0},
    {M_AFS_UPDATE,   "Update",           TRUE,   NULL,   0,0,0,0},
    {M_AFS_FILEINFO, "File Info",        TRUE,   NULL,   0,0,0,0},
    {M_AFS_LONGLIST, "Long Listing",     TRUE,   NULL,   0,0,0,0}
};
#define AFS_MENU_SIZE    (sizeof(AFS_menu) / sizeof(MENU))

MENULOC afs_menu_loc = {AFS_MENU_SIZE, 0,0,0,0,0};

MENU QEdit_menu[] = {
    /*  
  {M_QE_PREFS,   "preferences",    TRUE, NULL, 0,0,0,0},
  {M_QE_DOTPREFS,".preferences",   TRUE, NULL, 0,0,0,0}, */
    {M_QE_ALIASES, ".aliases",       TRUE, NULL, 0,0,0,0},
    {M_QE_SIGFILE, ".signature",     TRUE, NULL, 0,0,0,0}
};
#define QEDIT_MENU_SIZE  (sizeof(QEdit_menu) / sizeof(MENU))

MENULOC qedit_menu_loc = {QEDIT_MENU_SIZE, 0,0,0,0,0};

MENU Help_menu[] = {
#ifdef IBMPC
    {M_ABOUT,	    "PC-EzMail...",	TRUE,	NULL,	0,0,0,0},
#endif
#ifdef NEXT
    {M_ABOUT,	    "NeXT-EzMail...",	TRUE,	NULL,	0,0,0,0},
#endif
#ifdef VMS
    {M_ABOUT,	    "VMS-EzMail...",	TRUE,	NULL,	0,0,0,0},
#endif
    {M_ABOUT,	    "EzMail...",	TRUE,	NULL,	0,0,0,0},
    {M_BUG,	    "Bug Report",	TRUE,	NULL,	0,0,0,0},
    {M_HELP,	    "Help On EzMail",  	TRUE,	NULL,	0,0,KEYCODE_HELP,0},
    {M_KEYMAP,       "Help On Keys",     TRUE,   NULL,   0,0,KEYCODE_KEY_MAP,0},
    {NULL_MENU},
    {M_QUIT,	    "Quit",	        TRUE,	NULL,	0,0,KEYCODE_QUIT,0}
};
#define Help_menu_SIZE    (sizeof(Help_menu) / sizeof(MENU))

MENULOC Help_menu_loc = {Help_menu_SIZE, 0,0,0,0,0};
/*
MENU WinAttr_menu[] = {
};
#define WINATTR_MENU_SIZE (sizeof(WinAttr_menu) / sizeof(MENU))

MENULOC winattr_menu_loc = {WINATTR_MENU_SIZE, 0,0,0,0,0};
*/
MENU Prefs_menu[] = {
  {M_LOCAL_PREFS,   "Local Preferences...", TRUE, NULL, 0,0,KEYCODE_ALT_J,0},
  {M_REMOTE_PREFS,  "Network Preferences...", TRUE, NULL, 0,0,0},
  /*  {M_REMOTE_PREFS,  "Andrew Preferences...", FALSE, NULL, 0,0,0,0}, */
  /*  {M_SIGN_THIS_MAIL,"Sign My Mail",          TRUE, NULL, 0,0,0,0},*/
  {M_BCC_THIS_MAIL, "Blind Carbon Copy",     TRUE, NULL, 0,0,0,0}
};

#define PREFS_MENU_SIZE (sizeof(Prefs_menu) / sizeof(MENU))

MENULOC prefs_menu_loc = {PREFS_MENU_SIZE, 0,0,0,0,0};

MENU Hdr_menu[] = {
  {M_ADD_HEADERS,     "Mail Headers",    FALSE, NULL, 0,0,KEYCODE_ADD_HEADERS,0},
  {M_ADD_FAXHDRS,     "Fax Headers...",            FALSE, NULL, 0,0,0,0},
  {M_ADD_ACK,         "Acknowledgement...",        TRUE,  NULL, 0,0,0,0},
  {M_ADD_VOTE,        "Request a Vote...",         TRUE,  NULL, 0,0,0,0},
  {M_ADD_SUB_INVITE,  "Subscription Invitation...",TRUE,  NULL, 0,0,0,0},
  {M_ADD_DIST_INVITE, "Redistribution Request...", TRUE,  NULL, 0,0,0,0},
  {M_ADD_REPLY_TO,    "Reply To...",               TRUE,  NULL, 0,0,0,0}
};

#define HDR_MENU_SIZE (sizeof(Hdr_menu) / sizeof(MENU))

MENULOC hdr_menu_loc = {HDR_MENU_SIZE, 0,0,0,0,0};

MENU AllWindows_menu[] = {
    {M_SHOW_FOLDERS, "Folders",	      FALSE,NULL,0,0,KEYCODE_FOLDERS,0},
    {M_SHOW_CAPTIONS,"Captions",	      FALSE,NULL,0,0,KEYCODE_CAPTIONS,0},
    {M_SHOW_MESSAGE, "Message",	      FALSE,NULL,0,0,KEYCODE_MESSAGE,0},
    {M_SHOW_FILER,   "Andrew Filer",   FALSE,NULL,0,0,KEYCODE_FILER,0},
    {M_SHOW_CONSOLE, "Console",	      TRUE, NULL,0,0,KEYCODE_CONSOLE,0},
     {M_SHOW_DOCUMENT,"Document(s)",    TRUE, NULL,0,0,KEYCODE_DOCUMENT,0}
};

#define ALLWIN_MENU_SIZE	(sizeof(AllWindows_menu) / sizeof(MENU))

MENULOC allwin_menu_loc = {ALLWIN_MENU_SIZE, 0,0,0,0,0};

MENU Window_menu[] = {
    {M_BRING_FRONT,"Bring-To-Front",TRUE,AllWindows_menu,&allwin_menu_loc,0,0,0},
     {NULL_MENU},
    {M_MOVE,	  "Move",	TRUE,	NULL,	0,0,0,0},
    {M_RESIZE,	  "Resize",	TRUE,	NULL,	0,0,0,0},
    {M_ZOOM,       "Zoom",     TRUE,   NULL,   0,0,KEYCODE_ZOOM,0},
    {M_REDRAW,     "Redraw",   TRUE,   NULL,   0,0,KEYCODE_REDRAW,0},
     {NULL_MENU},
#ifdef AFS_ENV
    {M_SAVE,       "Save...",  TRUE,   NULL,       0,0,KEYCODE_LOCAL_SAVE,0},
    {M_SAVE_ATTACH, "Save Attachment", TRUE, NULL, 0, 0, KEYCODE_SAVE_ATTACH, 0 },
#else /* AFS_ENV */
    {M_SAVE,       "Save",     TRUE,	Save_menu,  &save_menu_loc,0,0,0},
    {M_PRINT_LOCAL,"Local Print",TRUE, NULL,0,0,KEYCODE_LOCAL_PRINT,0},
#endif /* AFS_ENV */
     {M_CLOSE,	  "Close Window",FALSE,NULL,       0,0,KEYCODE_CLOSE,0}
};
#define WINDOW_MENU_SIZE    (sizeof(Window_menu) / sizeof(MENU))

MENULOC window_menu_loc = {WINDOW_MENU_SIZE, 0,0,0,0,0};

MENU Edit_menu[] = {
  {M_EDITMSG,      "Internal Edit", FALSE, NULL, 0,0,KEYCODE_EDIT,0},
  {M_EXTERNAL_EDIT,"External Edit", FALSE, NULL, 0,0,0,0},
   {NULL_MENU},
  {M_SET_MARK,     "Set Mark",      TRUE,  NULL, 0,0,KEYCODE_SET_MARK,0},
  {M_CUT,          "Cut",           FALSE, NULL, 0,0,KEYCODE_CUT,0},
  {M_COPY,         "Copy",          FALSE, NULL, 0,0,KEYCODE_COPY,0},
  {M_PASTE,        "Paste",         FALSE, NULL, 0,0,KEYCODE_PASTE,0},
   {NULL_MENU},
   {M_PREFERENCES,  "Preferences",TRUE,Prefs_menu,&prefs_menu_loc,0,KEYCODE_ALT_J,0}
};

#define EDIT_MENU_SIZE  (sizeof(Edit_menu) / sizeof(MENU))

MENULOC Edit_menu_loc = {EDIT_MENU_SIZE, 0,0,0,0,0};

MENU Compose_menu[] = {
    {M_NEW,	    "New",	TRUE,	NULL,	    0,0,KEYCODE_COMPOSE,0},
    {M_MAIL_HEADERS, "Add Headers",FALSE,Hdr_menu,&hdr_menu_loc,0,KEYCODE_ADD_HEADERS,0},
#ifdef AFS_ENV
    {M_OPEN_LOCAL,   "Open File...",TRUE, NULL,     0,0,KEYCODE_LOCAL_OPEN,0},
    {M_APPEND,       "Append...",FALSE,  NULL,       0,0,0,0},
#else /* AFS_ENV */
    {M_OPEN,	    "Open File",TRUE,	Open_menu,  &open_menu_loc,0,0,0},
    {M_APPEND,	    "Append",   FALSE,	Append_menu,&append_menu_loc,0,0,0},
#endif /* AFS_ENV */
     {NULL_MENU},
    {M_SEND,	    "Send",     FALSE,  NULL,	0,0,KEYCODE_SEND,0},
    {M_VERIFY,	    "Verify",   FALSE,  NULL,	0,0,KEYCODE_VERIFY,0},
    /*{M_SIGNMSG,	    "Sign Message",   FALSE,  NULL,	0,0,0,0}*/
    /* xxx     {M_WHOIS,	    "Who is...",FALSE,  NULL,   0,0,0,0}*/
};
#define COMPOSE_MENU_SIZE	(sizeof(Compose_menu) / sizeof(MENU))

MENULOC Compose_menu_loc = {COMPOSE_MENU_SIZE, 0,0,0,0,0};

MENU Folders_menu[] = {
    {M_CREATE,	"Create...",	FALSE,	NULL,	    0,0,0,0},
    {M_DELETE,	"Delete...",	FALSE,	NULL,	    0,0,0,0},
    {M_RENAME,	"Rename...",	FALSE,	NULL,	    0,0,0,0},
    /* xxx {M_MERGE,	"Merge...",	FALSE,	NULL,	    0,0,0,0}, */
     {NULL_MENU},
    {M_EXPOSE,	"Expose",       FALSE,	Expose_menu,&expose_menu_loc,0,KEYCODE_EXPOSE_MENU,0},
    {M_SUBSCRIBE,"Subscribe",    FALSE,	Subscribe_menu,&subscribe_menu_loc,0,KEYCODE_SUBS_MENU,0},
     {NULL_MENU},
    {M_INFO,	"Info",		FALSE,	NULL,	    0,0,0,0},
    {M_SEEN,	"Mark As Seen",	FALSE,	NULL,	    0,0,0,0},
    {M_PUNT,     "Punt",         FALSE,  NULL,       0,0,KEYCODE_PUNT,0},
    {M_POST,	"Post",		FALSE,	NULL,	    0,0,0,0},
     {M_QUOTA,	"Show Quota",	FALSE,	NULL,	    0,0,0,0}
};
#define FOLDERS_MENU_SIZE   (sizeof(Folders_menu) / sizeof(MENU))

MENULOC folders_menu_loc = {FOLDERS_MENU_SIZE, 0,0,0,0,0};

MENU Message_menu[] = {
  {M_NEXT,	"Next Message", FALSE,	NULL,	    0,0,KEYCODE_READ_NEXT,0},
  {M_PREV,	"Prev Message",	FALSE,	NULL,	    0,0,KEYCODE_READ_PREV,0},
  {M_KILL,	"Delete & Next",FALSE,	NULL,	    0,0,KEYCODE_KILL_NEXT,0},
  {M_MDELETE,	"Delete",	FALSE,	NULL,	    0,0,KEYCODE_MSG_DELETE,0},
  {M_UNDELETE,	"Undelete",	FALSE,	NULL,	    0,0,KEYCODE_MSG_UNDELETE,0},
  {NULL_MENU},
  {M_RESPOND,	"Respond",      FALSE,	Respond_menu,&respond_menu_loc,0,KEYCODE_REPLY_MENU,0},
  {M_FILEINTO,	"File into",    FALSE,	FileInto_menu,&fileinto_menu_loc,0,KEYCODE_FMSG_MENU,0},
  {NULL_MENU},
  {M_SETNEW,	"Set new",	FALSE,	NULL,	    0,0,KEYCODE_SETNEW,0},
  {M_HEADERS,	"Full Headers",	FALSE,	NULL,	    0,0,KEYCODE_AT,0},
  {M_ROT_13,     "Rot-13",       FALSE,  NULL,       0,0,0,0},
  {M_EXPUNGE,    "Purge Deletions",FALSE,NULL,       0,0,0,0},
  {M_EXCERPT,     "Excerpt Body",  FALSE, NULL, 0,0,KEYCODE_EXCERPT,0},
  {M_SEARCH,	"Search", FALSE,	NULL,	    0,0,KEYCODE_S,0},
};
#define MESSAGE_MENU_SIZE   (sizeof(Message_menu) / sizeof(MENU))

MENULOC message_menu_loc = {MESSAGE_MENU_SIZE, 0,0,0,0,0};

#ifdef WANT_NULL_MENU
#define RESPOND_MENU_INDEX  6
#define FILEINTO_MENU_INDEX 7
#else
#define RESPOND_MENU_INDEX  5
#define FILEINTO_MENU_INDEX 6
#endif

MENU Services_menu[] = {
  {M_CONNECT,	  "Connect...",   TRUE, NULL,	0,0,0,0},
  {M_DISCONNECT,   "Disconnect",   FALSE,NULL,	0,0,0,0},
  {M_SUSPEND,    "Suspend",  TRUE, NULL, 0,0,KEYCODE_SUSPEND,0},
   {NULL_MENU},
  {M_OTHERSVC,     "Unix Services",FALSE,Andrew_menu,&andrew_menu_loc,0,KEYCODE_UNIX_SVC,0},
  {M_AFS_SERVICES, "File Services",FALSE,AFS_menu,   &afs_menu_loc,0,KEYCODE_AFS_SVC,0},
  {M_QUICK_EDIT,   "Quick Edit",   FALSE,QEdit_menu, &qedit_menu_loc,0,0,0},
  /*  {M_UPDATE,	  "Update Console",FALSE,NULL,0,0,0,0},*/
  /*{NULL_MENU},*/
  /*   {M_GETMAIL,	  "Get My Mail",  FALSE,NULL, 0,0,KEYCODE_GET_MAIL,0}*/
};
#define SERVICES_MENU_SIZE  (sizeof(Services_menu) / sizeof(MENU))

MENULOC services_menu_loc = {SERVICES_MENU_SIZE, 0,0,0,0,0};

MENU_TREE   menu_bar[] = {
  {M_EZMAIL,   Help_menu,      "Help",   TRUE,	&Help_menu_loc},
  {M_COMPOSE,  Compose_menu,   "Compose",TRUE,   &Compose_menu_loc},
  {M_EDIT,     Edit_menu,      "Edit",   TRUE,   &Edit_menu_loc},
  {M_FOLDERS,  Folders_menu,   "BBoards/Folders",FALSE,	&folders_menu_loc},
  {M_MESSAGE,  Message_menu,   "Message",FALSE,	&message_menu_loc},
  {M_SERVICES, Services_menu,  "Andrew", TRUE,	&services_menu_loc},
  {M_WINDOW,   Window_menu,    "Window",	TRUE,	&window_menu_loc}
};
#define	MENUBAR_SIZE	(sizeof(menu_bar) / sizeof(MENU_TREE))

void MNU_SetMenuCmdProc(int (*aProc)())
{
  if (aProc) MNU_DoMenuCommand = aProc;
}

MENU *MNU_FindMenu(int item)
{
  int i, j, k;
  MENU *menu, *submenu;

  for (i=0; i<MENUBAR_SIZE; i++) {
    for (j=0, menu=menu_bar[i].menu; 
	 j<menu_bar[i].loc->submenu_size; j++, menu++) {
      if (menu->id == item) return(menu);
      submenu = menu->submenu;
      if (submenu) {
      for (k=0; k<menu->loc->submenu_size; k++, submenu++)
	if (submenu->id == item) return(submenu);
      }
    }
  }
  return((MENU *)NULL);
}

char *MNU_BuildMenuName(char *name,int has_submenu, int key_code, int max_size)
{
  static char namebuf[STR255LEN];
  int len;

  strncpy(namebuf, name, max_size-COMMAND_KEY_WIDTH);
  namebuf[max_size - COMMAND_KEY_WIDTH] = '\0';
  len = strlen(namebuf);
  memset(&namebuf[len], (int)' ', imax(0, (max_size-len)));
  namebuf[max_size] = '\0';
  if (has_submenu) /* menus with submenus have no key_codes */
    strncpy(&namebuf[max_size-COMMAND_KEY_WIDTH+4], &MNU_HARROW[0], 2);
  else if (key_code != 0) {
    char *key_text = SCR_GetKeyText(key_code);
    strncpy(&namebuf[max_size-imin(COMMAND_KEY_WIDTH+1, strlen(key_text))],
	    key_text, COMMAND_KEY_WIDTH-1);
  }
  return(namebuf);
}

void ForEachMenuItem(void (*doThis)())
{
  int i, j, k;
  MENU *menu, *submenu;

  for (i=0; i<MENUBAR_SIZE; i++) {
    for (j=0, menu=menu_bar[i].menu; 
	 j<menu_bar[i].loc->submenu_size; j++, menu++) {
      (*doThis)(menu);
      submenu = menu->submenu;
      if (submenu) {
	  for (k=0; k<menu->loc->submenu_size; k++, submenu++)
	      (*doThis)(submenu);
      }
    }
  }
}

void InitEnableFlag(MENU *amenu)
{
  amenu->enabled = amenu->initial;
}

void MNU_RevertToInitial(void)
{
  ForEachMenuItem(InitEnableFlag);
}

void EnableAll(MENU *menu)
{
#ifdef WANT_NULL_MENU
  if (menu->id)
#endif
    menu->enabled = TRUE;
}

void DisableAll(MENU *menu)
{
  menu->enabled = FALSE;
}

void MNU_EnableAll(int aval)
{
  if (aval)
    ForEachMenuItem(EnableAll);
  else
    ForEachMenuItem(DisableAll);
}

void MNU_init_menuloc(int idx)
{
    int j, maxsize = 0;
    MENU *amenu;
    MENULOC *mloc = menu_bar[idx].loc;

    mloc->pos = (idx>0 ? 
	    menu_bar[idx-1].loc->pos+menu_bar[idx-1].loc->size : 0) + 2;
    mloc->size = strlen(menu_bar[idx].name) + 1;
    for (amenu=menu_bar[idx].menu,j=0; j<mloc->submenu_size; amenu++,j++) {
	maxsize = imax(strlen(amenu->name), maxsize);
	amenu->initial = amenu->enabled;
    }

    maxsize += COMMAND_KEY_WIDTH;
    mloc->box_width = maxsize + 2 * (1 + INDENT);
    mloc->box_height = mloc->submenu_size + 2;
    mloc->pop_row = menubar_row + 1;
    mloc->pop_col = (mloc->pos+mloc->box_width >= LINE_WIDTH ?
		     LINE_WIDTH - mloc->box_width - 1 : mloc->pos);
}

void MNU_init_submenuloc(int idx, int sidx)
{
    int j, maxsize = 0;
    MENU *asubmenu, *amenu = menu_bar[idx].menu + sidx;

    amenu->loc->pos = menu_bar[idx].loc->pos;
    amenu->loc->size = strlen(amenu->name);
    for (asubmenu=amenu->submenu,j=0; 
	    j<amenu->loc->submenu_size; asubmenu++,j++) {
	maxsize = imax(strlen(asubmenu->name), maxsize);
	asubmenu->initial = asubmenu->enabled;
    }

    maxsize += (COMMAND_KEY_WIDTH + 2 * (1 + INDENT));
    amenu->loc->box_height = amenu->loc->submenu_size + 2;
    amenu->loc->pop_row = menu_bar[idx].loc->pop_row + sidx;
    j = menu_bar[idx].loc->pop_col + menu_bar[idx].loc->box_width;

    if ((screen_cols - j) < maxsize) {
      if (menu_bar[idx].loc->pop_col > maxsize) {
	j = menu_bar[idx].loc->pop_col - maxsize;
      } else {
	if ((screen_cols-j) >= (menu_bar[idx].loc->pop_col))
	  maxsize = screen_cols - j;
	else {
	  j = 0;
	  maxsize = menu_bar[idx].loc->pop_col;
	}
      }
    }
    amenu->loc->box_width = maxsize;
    amenu->loc->pop_col = j;
}

void MNU_initialize(int (*do_menu_command_proc)())
{
    int i, j;
    MENU *amenu;

    MNU_SetMenuCmdProc(do_menu_command_proc);

    for (i=0; i<MENUBAR_SIZE; i++) {
	MNU_init_menuloc(i);
	for (amenu=menu_bar[i].menu,j=0;
	    j<menu_bar[i].loc->submenu_size; amenu++,j++)
	    if (amenu->submenu) MNU_init_submenuloc(i, j);
    }
}

void MNU_HighlightMenu(int on_off)
{
  if (menutree_idx >= 0) {
    MENULOC *amenuloc = menu_bar[menutree_idx].loc;
    SCR_Write(menu_bar[menutree_idx].name, menubar_row,
	      amenuloc->pos+1, amenuloc->size-1, on_off);
  }
}

void highlight_item(MENU *amenu,MENULOC *amenuloc,int on_off, int idx)
{
  char *nameptr;

  nameptr = MNU_BuildMenuName(amenu->name, 
			      amenu->submenu ? TRUE : FALSE,
			      amenu->bykey,
			      amenuloc->box_width-2 * (1 + INDENT));
  SCR_Write(nameptr, 
	    amenuloc->pop_row + idx + TEXT_ROW, 
	    amenuloc->pop_col + TEXT_COL,
	    amenuloc->box_width - 2 * (1 + INDENT), on_off);

  SCR_UpdateScreen();
}

void MNU_HighlightHItem(int on_off)
{
  if (menuharc_idx >= 0) {
    MENU *amenu = menu_bar[menutree_idx].menu + menuitem_idx;
    highlight_item((amenu->submenu + menuharc_idx), amenu->loc, 
		   on_off, menuharc_idx);
  }
}

void MNU_HighlightItem(int on_off)
{
  if (menuitem_idx >= 0) {
    MENU *amenu = menu_bar[menutree_idx].menu + menuitem_idx;
    highlight_item(amenu, menu_bar[menutree_idx].loc, on_off, menuitem_idx);
    if (amenu->submenu) {
      if (on_off == H_ON)
	MNU_PullDownHMenu(amenu);
      else
	MNU_CloseMenu(FALSE);
    }
  }
}

void MNU_RedrawBarItem(int item)
{
  int attr;
  char disabled = SPC_KEY;

  if (menu_bar[item].enabled) attr = H_OFF;
  else {
    attr = H_DIM;
    if (SCR_CanHighlight() == FALSE) disabled = MNU_DISABLED_CHAR;
  }

  SCR_Write(&disabled, menubar_row, menu_bar[item].loc->pos, 1, attr);

  SCR_Write(menu_bar[item].name, menubar_row,
	    menu_bar[item].loc->pos+1, menu_bar[item].loc->size-1, attr);
}

void MNU_SetupMenuBar(void)
{
  int i;

  for (i=0; i<MENUBAR_SIZE; i++) MNU_RedrawBarItem(i);

  SCR_UpdateScreen();
}

void MNU_DisplayMenu(int which)
{
  if (menu_isopen) {
    MNU_CloseMenu(TRUE);
    menuitem_idx = -1;
    menuharc_idx = -1;
  }
  MNU_HighlightMenu(H_OFF);
  menutree_idx = which;
  MNU_HighlightMenu(H_ON);
  SCR_UpdateScreen();
  MNU_PullDownMenu();
}

int MNU_GotoNextMenuBar(void)
{
  int i;

  for (i=WRAP_INDEX(menutree_idx, MENUBAR_SIZE);
       i<MENUBAR_SIZE && i!=menutree_idx;
       i = (menutree_idx < 0 ? (i+1) : WRAP_INDEX(i, MENUBAR_SIZE)))
    if (menu_bar[i].enabled) {
      MNU_DisplayMenu(i);
      break;
    }

  menuitem_idx = -1;
  menuharc_idx = -1;
  return(TRUE);
}

int MNU_GotoPrevMenuBar(void)
{
  int i;

  for (i=BWRAP_INDEX(menutree_idx, MENUBAR_SIZE);
       i<MENUBAR_SIZE && i!=menutree_idx;
       i = (menutree_idx < 0 ? (i+1) : BWRAP_INDEX(i, MENUBAR_SIZE)))
    if (menu_bar[i].enabled) {
      MNU_DisplayMenu(i);
      break;
    }

  menuitem_idx = -1;
  menuharc_idx = -1;
  return(TRUE);
}

int MNU_GetPrevEnabled(int this, int size,MENU *menu)
{
  int i, j;
  MENU *menu_head = menu;

  if (this < 0) this = -1;
  menu = menu + BWRAP_INDEX(this, size);
  for (i=BWRAP_INDEX(this,size), j=0; i!=this && j<size;
       menu = menu_head + BWRAP_INDEX(i, size),i=BWRAP_INDEX(i, size), j++)
    if (menu->enabled) return(i);

  return(this);
}

int MNU_GetNextEnabled(int this, int size,MENU *menu)
{
  int i, j;
  MENU *menu_head = menu;

  if (this >= 0) menu = menu + WRAP_INDEX(this, size);
  for (i=WRAP_INDEX(this,size), j=0; i!=this && j<size;
       menu = menu_head + WRAP_INDEX(i, size),i=WRAP_INDEX(i, size), j++)
    if (menu->enabled) return(i);

  return(this);
}

int MNU_MoveUpDown(int which)
{
  int (*getidx)();

  if (which == KEYCODE_DOWN)
    getidx = MNU_GetNextEnabled;
  else if (which == KEYCODE_UP)
    getidx = MNU_GetPrevEnabled;
  else return(FALSE);

  if (menuharc_idx < 0)
    MNU_HighlightItem(H_OFF);
  else
    MNU_HighlightHItem(H_OFF);

  if (menuharc_idx < 0)
    menuitem_idx = (*getidx)(menuitem_idx, 
			       menu_bar[menutree_idx].loc->submenu_size, 
			       menu_bar[menutree_idx].menu);
  else
    menuharc_idx = (*getidx)(menuharc_idx,
	menu_bar[menutree_idx].menu[menuitem_idx].loc->submenu_size,
        menu_bar[menutree_idx].menu[menuitem_idx].submenu);

  if (menuharc_idx < 0)
    MNU_HighlightItem(H_ON);
  else
    MNU_HighlightHItem(H_ON);

  SCR_UpdateScreen();
  return(TRUE);
}

int MNU_MoveLeftRight(int which)
{
  if (menuitem_idx >= 0 && menu_bar[menutree_idx].menu[menuitem_idx].submenu) {
    if (menuharc_idx < 0) {
      menuharc_idx = MNU_GetNextEnabled(menuharc_idx,
		menu_bar[menutree_idx].menu[menuitem_idx].loc->submenu_size,
		menu_bar[menutree_idx].menu[menuitem_idx].submenu);
      MNU_HighlightHItem(H_ON);
    } else {
      MNU_HighlightHItem(H_OFF);
      menuharc_idx = -1;
    }
    SCR_UpdateScreen();
    return(TRUE);
  }
  return(FALSE);
}

int search_menu(MENU *amenu,MENULOC *amenuloc,int start_idx, int keycode)
{
  MENU *first_menu = amenu;
  int i, j, cur_idx, upper_key = SCR_UpCaseRepresentation(keycode);

  for (j=0; j<start_idx; j++) amenu++;
  j = cur_idx = start_idx;

  for (i=0; i<amenuloc->submenu_size; i++) {
    j = WRAP_INDEX(j,amenuloc->submenu_size);
    if (j == 0) amenu = first_menu;
    else amenu++;
    if (j != cur_idx && amenu->enabled) {
      char ch = amenu->name[0];
      if (ch == keycode || ch == upper_key) {
	cur_idx = j;
	break;
      }
    }
  }
  return(cur_idx == start_idx ? -1 : cur_idx);
}

int MNU_GotoMenuByInitial(int keycode)
{
  int j, cur_idx;
  MENU *amenu = menu_bar[menutree_idx].menu;
  MENULOC *amenuloc;

  for (j=0; j<menuitem_idx; j++) amenu++;

  if (menuitem_idx >= 0 && amenu->submenu) {
    cur_idx = search_menu(amenu->submenu, amenu->loc, menuharc_idx, keycode);
    if (cur_idx >= 0 && menuharc_idx < 0) menuharc_idx = cur_idx;
  } else
    cur_idx = -1;

  if (cur_idx < 0) {
    amenuloc = menu_bar[menutree_idx].loc;
    amenu = menu_bar[menutree_idx].menu;
    cur_idx = search_menu(amenu, amenuloc, menuitem_idx, keycode);
    if (cur_idx >= 0 && menuharc_idx >= 0) {
      MNU_HighlightHItem(H_OFF);
      menuharc_idx = -1;
    }
  }

  if (cur_idx >= 0) {
    if (menuharc_idx >= 0) {
      MNU_HighlightHItem(H_OFF);
      menuharc_idx = cur_idx;
      MNU_HighlightHItem(H_ON);
    } else {
      MNU_HighlightItem(H_OFF);
      menuitem_idx = cur_idx;
      MNU_HighlightItem(H_ON);
    }
    return(TRUE);
  }
  return(FALSE);
}

int MNU_ProcessKey(int keycode)
{
  int i, key;

  /* JCS - if you leave a menu open, for some reason, you eventually
   * get this keycode even though no key has been pressed.  Silently ignore.
   */
  if ( keycode == 127 ) { 
    return (TRUE);
  }

  if (menuitem_idx < 0 || 
      menu_bar[menutree_idx].menu[menuitem_idx].submenu == (MENU *)NULL) {
    if (keycode == KEYCODE_RIGHT) return(MNU_GotoNextMenuBar());
    else if (keycode == KEYCODE_LEFT) return(MNU_GotoPrevMenuBar());
  }

  if (menu_isopen) {
    if (keycode == KEYCODE_UP || keycode == KEYCODE_DOWN)
      return(MNU_MoveUpDown(keycode));
    else if (keycode == KEYCODE_RIGHT || keycode == KEYCODE_LEFT)
      return(MNU_MoveLeftRight(keycode));
  }

  if (menu_isopen) {
    int save_menutree = menutree_idx;
    int save_menuitem = menuitem_idx;
    int save_menuharc = menuharc_idx;
    if (isascii(keycode) && isprint(keycode)) {
      if (MNU_GotoMenuByInitial(keycode)) return(TRUE);
    } else {
      int menuid = MNU_FindMenuByKey(keycode); /* look for key mapping */
      if (menuid > 0) {
	int new_tree = menutree_idx;
	int new_item = menuitem_idx;
	menutree_idx = save_menutree;
	menuitem_idx = save_menuitem;
	menuharc_idx = save_menuharc;
	if (menutree_idx == new_tree) { /* in the same menu */
	  if (new_item >= 0 && menuitem_idx == new_item) return(TRUE);
	  if (menuitem_idx >= 0) MNU_HighlightItem(H_OFF);
	  menuitem_idx = new_item;
	  if (menuitem_idx >= 0) {
	    MNU_HighlightItem(H_ON);
	    MNU_ProcessKey(KEYCODE_RIGHT);
	  } else
	    MNU_ReturnControl(menuid);
	  return(TRUE);
	} else { /* in another menu */
	  MNU_ReturnControl(0);
	  menutree_idx = new_tree;
	  menuitem_idx = new_item;
	  if (menuitem_idx >= 0) {
	    MNU_PullDownMenu();
	    MNU_HighlightItem(H_ON);
	    MNU_ProcessKey(KEYCODE_RIGHT);
	  } else
	    MNU_ReturnControl(menuid);
	  return(TRUE);
	}
      }
    }
    menutree_idx = save_menutree;
    menuitem_idx = save_menuitem;
    menuharc_idx = save_menuharc;
  }

  key = (isascii(keycode)&&islower(keycode)? toupper(keycode): keycode);

  for (i=WRAP_INDEX(menutree_idx, MENUBAR_SIZE);
       i<MENUBAR_SIZE && i!=menutree_idx;
       i = (menutree_idx < 0 ? (i+1) : WRAP_INDEX(i, MENUBAR_SIZE)))
    if (key == menu_bar[i].name[0] && menu_bar[i].enabled) {
      MNU_DisplayMenu(i);
      return(TRUE);
    }

  return(FALSE);
}

void MNU_ReturnControl(int do_cmd)
{
  if (menu_isopen && menu_bar[menutree_idx].enabled) MNU_CloseMenu(TRUE);

  if (do_cmd != 0) {
    (*MNU_DoMenuCommand)(do_cmd);
    lastmenu_cmd = do_cmd;
    lastmenu_idx = menutree_idx;
  } else {
    lastmenu_cmd = -1;
    lastmenu_idx = -1;
  }

  MNU_HighlightMenu(H_OFF);
  menu_isopen = FALSE;
  menuitem_idx = -1;
  menuharc_idx = -1;
  menutree_idx = -1;
  SCR_UpdateScreen();
}

void MNU_MenuMode(int keycode)
{
  menu_focus = FALSE;
  harc_focus = FALSE;

  if (keycode == KEYCODE_SHIFT_TAB) {
    MENU *amenu;
    if (lastmenu_cmd > 0 && lastmenu_idx >= 0 && 
	((amenu = MNU_FindMenu(lastmenu_cmd))) && amenu->enabled) {
      menutree_idx = lastmenu_idx;
      MNU_HighlightMenu(H_ON);
      MNU_ReturnControl(lastmenu_cmd);
    } else { 
      SCR_ClearKeyboard();
      SCR_Beep();
    }
  }
  else if (keycode == TAB_KEY)
    MNU_SelectMenuBar(keycode);
  else {
    int mid = MNU_FindMenuByKey(keycode);
    if (mid) {
      MNU_HighlightMenu(H_ON);
      if (menuitem_idx < 0)
	MNU_ReturnControl(mid);
      else { /* open menu, and submenu */
	MNU_PullDownMenu();
	MNU_HighlightItem(H_ON);
	MNU_SelectMenuBar(KEYCODE_RIGHT);
      }
    } else
      MNU_SelectMenuBar(keycode); /* maybe the first letter of a menu */
  }
  WIN_HomeCursor();
}

void MNU_SelectMenuBar(int keycode)
{
  /*
    Once menu bar is selected, we stay in this routine
    until the function is performed and menu bar is released!
  */

  do {
    if (keycode == TAB_KEY) 
      MNU_GotoNextMenuBar();
    else if ((keycode == KEYCODE_RETURN) || (keycode == KEYCODE_FAKERETURN)) {
      if (menuitem_idx >= 0) {
	int   cmd_id;
	MENU *amenu = menu_bar[menutree_idx].menu + menuitem_idx;
	if (menuharc_idx < 0 && amenu->submenu) {
	  SCR_ClearKeyboard();
	  SCR_Beep();
	} else {
	  if (menuharc_idx < 0) cmd_id = amenu->id;
	  else cmd_id = (amenu->submenu + menuharc_idx)->id;
	  MNU_ReturnControl(cmd_id);
	}
      }
    }
    else if (keycode == menu_closekey || keycode == menu_alt_closekey)
      MNU_ReturnControl(0);
    else {
      if (!MNU_ProcessKey(keycode)) {
	SCR_ClearKeyboard();
	SCR_Beep();
      }
    }

    if (menutree_idx >= 0) keycode = SCR_GetKey();
  } while (menutree_idx >= 0);
}

void MNU_PullDownHMenu(MENU *amenu)
{
    pull_it_down(amenu->submenu, amenu->loc, &harc_focus);
}

void MNU_PullDownMenu(void)
{
  if (!menu_bar[menutree_idx].enabled) return;

  pull_it_down(menu_bar[menutree_idx].menu, 
	       menu_bar[menutree_idx].loc, &menu_focus);

  menu_isopen = TRUE;
}

void pull_it_down(MENU *amenu,MENULOC *amenuloc,Boolean *focus)
{
  int i;
  char *nameptr;

  MNU_DrawMenuBox(amenuloc, focus);

  for (i=0; i<amenuloc->submenu_size; amenu++, i++) {
    char chk_mark;
    int attr = (amenu->enabled ? H_OFF : H_DIM);
    if (amenu->check) chk_mark = MNU_CHKMRK[0];
    else chk_mark = SPC_KEY;
    if (!amenu->enabled && SCR_CanHighlight() == FALSE)
      chk_mark = MNU_DISABLED_CHAR;
    SCR_Write(&chk_mark, amenuloc->pop_row + i + TEXT_ROW,
	      amenuloc->pop_col + FLAG_COL, 1, attr);

#ifdef WANT_NULL_MENU
    if (amenu->id == 0) {
#ifndef IBMPC
      char dash_buf[STR255LEN];
      memset(dash_buf, '-', sizeof(dash_buf));
      SCR_Write(dash_buf, 
		amenuloc->pop_row + i + TEXT_ROW, amenuloc->pop_col + 1,
		amenuloc->box_width - 2, NORMAL);
#else /* IBMPC */
      SCR_Line(1, amenuloc->box_width - 2, amenuloc->pop_row + i + TEXT_ROW,
	       amenuloc->pop_col + 1,
	       MENU_BORDER_ATTRIBUTE, 0L, SINGLE_LINE_BOX);
#endif /* IBMPC */
    } else
#endif /* WANT_NULL_MENU */
      {
      nameptr = MNU_BuildMenuName(amenu->name, (amenu->submenu ? TRUE : FALSE),
				  amenu->bykey,
				  amenuloc->box_width-2 * (1 + INDENT));
      SCR_Write(nameptr, amenuloc->pop_row + i + TEXT_ROW,
		amenuloc->pop_col + TEXT_COL,
		amenuloc->box_width-2 * (1 + INDENT) + 1, attr);
    }
  }

  SCR_SetCursorPosn(0, 0);
  SCR_UpdateScreen();
}

void MNU_DrawMenuBox(MENULOC *amenuloc,Boolean *focus)
{
  SCR_Box(amenuloc->box_height, 
	  amenuloc->box_width, 
	  amenuloc->pop_row,
	  amenuloc->pop_col, MENU_BORDER_ATTRIBUTE, 0L);

  *focus = TRUE;
}

void MNU_CloseMenu(int all)
{
  int r =0;
  int c = 0;
  int h = 0;
  int w = 0;
  MENULOC *amenuloc;

  if (harc_focus) {
    amenuloc = menu_bar[menutree_idx].menu[menuitem_idx].loc;
    r = amenuloc->pop_row; c = amenuloc->pop_col;
    h = amenuloc->box_height; w = amenuloc->box_width;
  }

  if (all || (!harc_focus && menu_focus)) {
    amenuloc = menu_bar[menutree_idx].loc;
    menu_focus = FALSE;
    if (harc_focus) {
      h = imax(r+h, amenuloc->pop_row+amenuloc->box_height);
      w = imax(c+w, amenuloc->pop_col+amenuloc->box_width);
      r = imin(r, amenuloc->pop_row);
      c = imin(c, amenuloc->pop_col);
      h -= r;
      w -= c;
    } else {
      r = amenuloc->pop_row; c = amenuloc->pop_col;
      h = amenuloc->box_height; w = amenuloc->box_width;
    }
  }

  harc_focus = FALSE;

  if (!harc_focus && !menu_focus) menu_isopen = FALSE;

  WIN_RefreshRegion(r, c, h, w);
  SCR_SetCursorPosn(0, 0);
  SCR_UpdateScreen();
}

void MNU_EnableItem(int item, int aval)
{
  MENU *menu;

  menu = MNU_FindMenu(item);
  if (menu)
    menu->enabled = aval;
}

static int remote_fileinto_id = REMOTE_FILEINTO;

void MNU_ResetFileInto(void)
{
  MENU *menu = MNU_FindMenu(M_FILEINTO);

  if (menu) {
    if (alt_fileinto_menu) {
      int i;
      for (i=fileinto_menu_loc.submenu_size; 
	   i<alt_fileinto_menu_loc.submenu_size; i++)
	if ((alt_fileinto_menu+i)->name) free((alt_fileinto_menu+i)->name);
      free(alt_fileinto_menu);
    }
    alt_fileinto_menu = NULL;
    menu->submenu = FileInto_menu;
    menu->loc = &fileinto_menu_loc;
    remote_fileinto_id = REMOTE_FILEINTO;
  }
}

void MNU_AddFileinto(char *name,
		     int add_or_remove) /* add == +1, remove = -1 */
{
  MENU *submenu, *menu = MNU_FindMenu(M_FILEINTO);
  int i, bsize = sizeof(MENU) * menu->loc->submenu_size;

  if (alt_fileinto_menu == NULL) {
    if (add_or_remove < 0) return; /* How can we remove from nothing! */
    alt_fileinto_menu = (MENU *)malloc(bsize);
    if (alt_fileinto_menu == (MENU *)NULL) return;
    bcopy(menu->submenu, alt_fileinto_menu, bsize);
    bcopy(&fileinto_menu_loc, &alt_fileinto_menu_loc, sizeof(MENULOC));
    menu->submenu = alt_fileinto_menu;
    menu->loc = &alt_fileinto_menu_loc;
  } else {
    int i = 0;
    submenu = menu->submenu;
    while (++i <= menu->loc->submenu_size) {
      if (strcmp(submenu->name, name) == 0) {
	if (add_or_remove < 0) {
	  if (submenu->name) free(submenu->name);
	  if (i < menu->loc->submenu_size) { /* not last one, push up */
	    MENU *from = submenu;
	    bcopy(++from, submenu, (menu->loc->submenu_size-i)*sizeof(MENU));
	  }
	  menu->loc->submenu_size--;
	  goto AddFileinto_DONE; /* don't realloc when we remove */
	} else
	  return; /* already there, don't add it again */
      }
      submenu++;
    }
  }

  if (add_or_remove < 0) return; /* didn't find one to remove */

  alt_fileinto_menu = (MENU *)realloc(alt_fileinto_menu,
				      bsize + (add_or_remove * sizeof(MENU)));

  if (alt_fileinto_menu == (MENU *)NULL) {
    alt_fileinto_menu = menu->submenu; /* restore it */
    return;
  } else {
    menu->submenu = alt_fileinto_menu;
  }

  if (add_or_remove >= 0) {
    for (submenu=menu->submenu, i=0; i<menu->loc->submenu_size; i++)
      submenu++;
    bzero(submenu, sizeof(MENU));
    submenu->name = (char *)malloc(strlen(name)+1);
    if (submenu->name == (char *)NULL) return;
    strcpy(submenu->name, name);
    submenu->id = remote_fileinto_id--;
    submenu->enabled = TRUE;
    menu->loc->submenu_size++;
  }

AddFileinto_DONE:
  MNU_init_submenuloc(M_MESSAGE-1, FILEINTO_MENU_INDEX);
  MNU_init_submenuloc(M_MESSAGE-1, RESPOND_MENU_INDEX);
}

void MNU_EnableBar(int bar_item, int aval)
{
  int i;

  for (i=0; i<MENUBAR_SIZE; i++)
    if (menu_bar[i].id == bar_item) {
      menu_bar[i].enabled = aval;
      MNU_RedrawBarItem(i);
      return;
    }
}

int MNU_FindMenuByKey(int keycode)
{
  int i, j, k, upper_key;
  MENU *amenu, *submenu;

  menutree_idx = menuitem_idx = menuharc_idx = -1;

  upper_key = SCR_UpCaseRepresentation(keycode);
  for (i=0; i<MENUBAR_SIZE; i++) {
    if (menu_bar[i].enabled == FALSE) continue;
    for (amenu=menu_bar[i].menu,j=0;
	 j<menu_bar[i].loc->submenu_size; amenu++,j++) {
      if (amenu->bykey == keycode || amenu->bykey == upper_key) {
	if (amenu->enabled) {
	  if (amenu->submenu) menuitem_idx = j;
	  menutree_idx = i;
	  return(amenu->id);
	} else return(0);
      }
      if (amenu->submenu)
	for (submenu=amenu->submenu, k=0; k<amenu->loc->submenu_size;
	     k++, submenu++) {
	  if (submenu->bykey == keycode || submenu->bykey == upper_key) {
	    if (submenu->enabled) {
	      menutree_idx = i;
	      return(submenu->id);
	    } else return(0);
	  }
        }
    }
  }
  return(0);
}

void MNU_GetMenuName(int mid,char *astr)
{
  MENU *amenu = MNU_FindMenu(mid);

  if (amenu) strcpy(astr, amenu->name);
  else *astr = 0;
}

void MNU_CheckItem(int item, int on_off)
{
  MENU *amenu = MNU_FindMenu(item);

  if (amenu) amenu->check = on_off;
}

void MNU_SetKeyCode(int item, int keycode)
{
  MENU *amenu = MNU_FindMenu(item);

  if (amenu) amenu->bykey = keycode;
}

