/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "common.h"
#ifndef _MIME_H_
#define _MIME_H_ 

int rfc2045_from_mime_quoted_printable ( const char * in, 
					 char * out,
					 int    outLen );

int rfc2045_to_mime_quoted_printable ( const char * s,
				       char       * d,
				       int          outLen );

char * rfc2045_mime_headers ( char * type,
			      char * subtype,
			      char * charset,
			      char * encoding,
			      char * buffer,
			      int    bufLen );

#define MIME_TYPE_INVALID     -1
#define MIME_TYPE_TEXT        0
#define MIME_TYPE_IMAGE       1
#define MIME_TYPE_AUDIO       2
#define MIME_TYPE_VIDEO       3
#define MIME_TYPE_APPLICATION 4
#define MIME_TYPE_EXTENSION   5
#define MIME_TYPE_MESSAGE     6
#define MIME_TYPE_MULTIPART   7

typedef struct MimeAttachment_t
{
  int  type;
  char subtype[STR255LEN];
  char parameter[STR255LEN];
  char id[STR255LEN];
  char description[STR255LEN];
  char encoding[STR255LEN];
  int  sizeOctets;
  int  sizeLines;
  char ext_parameter[STR255LEN];
  char ext_disposition[STR255LEN];
  char ext_language[STR255LEN];
} MimeAttachment;

typedef struct MimeAttachmentList_t
{
  MimeAttachment       * value;
  struct MimeAttachmentList_t * leaf;
  struct MimeAttachmentList_t * next;
} MimeAttachmentList;

MimeAttachmentList * rfc2060_decode_bodystructure ( char * bs );

#endif /* _MIME_H_ */
