/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <mmspfio.h>

#include <stdio.h>
#include <sys/param.h>

#define v stdout
#define woserr(xx_arg) {fprintf xx_arg;perror(" - ");fflush(v);}
#define werr(xx_arg) {fprintf xx_arg;fflush(v);}
/* temporary error macro */
#define Fatal(xx_cond,xx_prx) \
 {if((xx_cond)!=0) { \
  fprintf xx_prx ;\
  perror(" - "); \
  exit(22); \
 }}

/*
 * linked list of all the blocks in a pool
 */
struct pool_header_R {
  struct pool_header_R *next;
};
typedef struct pool_header_R pool_header,*pool_header_pt;

pool_header_pt pl_malloc();

struct pool_R {
  pool_header h;
  char *sblock;			/*current string block*/
  long remaining;		/*bytes remaining in sblock*/
};
typedef struct pool_R pool,*pool_pt;
#define POOL_block_size (3500)

#define RL_fl_new_serial (1)
#define RL_fl_new_size (2)

struct RLS_header_R {
  char **data;			/*pointer to table of pointers*/
  long pt_size;			/*number of entries pointed to by page table*/
  pool lp;
};
typedef struct RLS_header_R RLS_header;
#define SERV_OR_CLIENT_FIELDS RLS_header s;

#include <rli.h>

/*
 * argument block passed to list clients on
 * incomming messages
 */
struct RLA_R {
  long lindex;			/*selected index*/
  char str[2*MAXPATHLEN+1];	/*argument string*/
};
typedef struct RLA_R RLA,*RLA_pt;

void RL_xnew();
void RL_xresize();
void RL_xappend();
int RL_read();

void RL_patch();
void RL_replace();

/*maybe non 4.3 system?*/
#ifndef FD_SET_SIZE
#define FD_SET_SIZE (8*sizeof(int))
#endif

#define er_ucall(xx_ures,xx_err){if((xx_ures)<0){rl_error xx_err ;};}
#define assert(xx_flag,xx_ldata) {if(!(xx_flag)){rl_error xx_ldata;}}
#define CKCUI(xx_fun,xx_err)\
  {if((mserrcode=(xx_fun))!=0){rl_cui_error xx_err;};}

#ifndef IBMPC
extern int errno;
#endif /* IBMPC */
