#include "charset.h"
#include <stdio.h>
#include <unistd.h>

char inbuffer[10000] = { 0 };
char outbuffer[30000] = { 0 };

int demimeqp ( const char * in, 
	       char       * out,
	       int          outLen );

int mimeqp ( const char * in,
	     char       * out,
	     int          outLen );


const char HEX_CHARS[]             = "0123456789ABCDEF";
const char RFC2045_SPECIAL_CHARS[] = "@.,;:<>[]\\\"()?/= \t";
#define    RFC2045_MAX_LINE_LEN 75

int rfc2045_is_safe_char ( const char c )
{
  return ( ( ( c >= 33 && c <= 126 ) && 
	     ( c != 61 ) && 
	     ( !strchr ( RFC2045_SPECIAL_CHARS, c ) ) ) ? 1 : 0 );
}

int rfc2045_is_safe_ws ( const char c ) 
{
  return ( ( c == ' ' || c == '\t' ) ? 1 : 0 );
}

char rfc2045_hex_to_char ( const char * value )
{
  char
    buf[2];
  
  buf[0] = value[0];
  buf[1] = value[1];

  return (char) strtol ( buf, NULL, 16 );
}

int rfc2045_from_mime_quoted_printable ( const char * in, 
					 char * out,
					 int    outLen )
{
  char 
    * outP,
    * inP;
  
  int
    outPos = 0;
  
  inP  = (char*)in;
  outP = (char*)out;
  
  if ( in == NULL || out == NULL || outLen <= 0 ) 
    return -1;
  
  while ( outPos < outLen && *inP != '\0' ) { 
    while ( *inP != '\0' && *inP != '=' && outPos < outLen && *inP != '\r' ) 
      *outP++ = *inP++, outPos++;

    /* Get rid of trailing whitespace */
    if ( *inP == '\r' ) { 
      char 
	* inP2 = inP - 1;

      while ( isspace ( *inP2-- ) )
	outP--, outLen--;
      inP += 2;
      *outP++ = '\r';
      *outP++ = '\n';
      outLen += 2;

    }

    if ( *inP == '=' ) {
      inP++;
      if ( *(inP) == '\0'|| *(inP+1) == '\0' ) {
	*outP = '\0';
	return outPos;
      }
      
      /* Check for an = at the end of the line */
      if ( !(*inP == '\r' && *(inP+1) == '\n') ) {
	*outP++ = rfc2045_hex_to_char ( inP );
	outPos++;
      }
      inP += 2;
    }
  }

  return outPos;
}


/* Need to have a buffer of at least three characters */
char * rfc2045_char_to_hex ( const char c, char * buf )
{
  buf[0] = HEX_CHARS[c / 16];
  buf[1] = HEX_CHARS[c % 16];
  buf[2] = '\0';
  return buf + 2;
}

/* This function is an RFC2045-compliant quoted-printable
 * encoder.  It takes a source buffer (null terminated) and 
 * destination buffer, along with the destination buffer's
 * length.
 * 
 * If there is enough buffer space, the length of the destination
 * buffer is returned, otherwise -1 is returned and the function
 * should be called with a larger buffer
 *
 */
int rfc2045_to_mime_quoted_printable ( const char * s,
				       char       * d,
				       int          outLen )
{
  char 
    * inP,
    * outP;

  int
    linePos = 0,
    outPos  = 0;
  
  inP  = (char*)s;
  outP = (char*)d;

  while ( *inP != '\0' ) { 
    
    if ( outPos + 1 >= outLen )
      return -1;

    if ( rfc2045_is_safe_char ( *inP ) ||
	 ( linePos < RFC2045_MAX_LINE_LEN && rfc2045_is_safe_ws ( *inP ) ) ) { 
      *outP++ = *inP++, outPos++;
      linePos++;

      if ( linePos >= RFC2045_MAX_LINE_LEN ) {
	if ( outPos + 3 >= outLen ) 
	  return -1;
	*outP++ = '=', outPos++;
	*outP++ = '\r', outPos++;
	*outP++ = '\n', outPos++;
	linePos = 0;
      }
      
    } else { 

      if ( outPos + 3 >= outLen )
	return -1;

      /* Need to encode it as a =XX character */
      if ( linePos >= ( RFC2045_MAX_LINE_LEN - 2 ) ) {
	*outP++ = '=', outPos++;
	*outP++ = '\r', outPos++;
	*outP++ = '\n', outPos++;
	linePos = 0;
      }
      
      *outP++ = '=', outPos++;
      outP = rfc2045_char_to_hex ( *inP++, outP ), outPos += 2;
      linePos += 3;
    }
  }

  *outP = '\0';
  return outPos;
}

int
main ( int argc, char ** argv )
{
  char * in;
  int i;

  if ( argc != 2 ) {
    printf ( "usage: %s u or e\n", argv[0] );
    return 0;
  }
  
  in = inbuffer;
  while ( ( *in = getchar() ) != EOF )
    in++;
  *in = 0;
  
  printf ( "I read: %d bytes\n", strlen ( inbuffer ) );
  
  if ( argv[1][0] == 'u' ) 
    rfc2045_from_mime_quoted_printable ( inbuffer, outbuffer, 29999 );
  else if ( argv[1][0] == 'e' ) 
    rfc2045_to_mime_quoted_printable ( inbuffer, outbuffer, 29999 );
  else {
    printf ( "Bad argument\n" );
    return 1;
  }
    

  printf ( "Decoded %d characters\n", strlen ( outbuffer ) );
  printf ( "%s", outbuffer );

  return 0;
}
